from musiclang import Chord, Element, Tonality
from musiclang.library import *

def test_chord_to_scale_pitches():
    chord = Element(2) % Tonality(1, mode="m", octave=1)
    scale_pitches = chord.scale_pitches

    assert scale_pitches == [16, 18, 20, 21, 24, 25, 27]


def test_initial_chord_to_scale_pitches():
    chord = Element(1)
    scale_pitches = chord.scale_pitches

    assert scale_pitches == [2, 4, 5, 7, 9, 11, 12]


def test_chord_with_mode_to_scale_pitches():
    chord = Element(1) % Tonality(1, mode="locrian", octave=0)
    scale_pitches = chord.scale_pitches

    assert scale_pitches == [2, 4, 6, 7, 9, 11, 13]

def test_chord_with_mode_to_scale_pitches2():
    T = Tonality(1, mode="locrian", octave=0)
    chord = Element(1) % (T + T)
    scale_pitches = chord.scale_pitches
    assert scale_pitches == [3, 5, 7, 8, 10, 12, 14]


def test_chord_pitches_seventh():

    chord = Element(0) % Tonality(0)
    assert chord['7'].chord_pitches == [0, 4, 7, 11]


def test_chord_pitches_ninth():

    chord = Element(0) % Tonality(0)
    assert chord['9'].chord_pitches == [0, 4, 7, 11, 14]

def test_chord_pitches():

    chord = Element(0) % Tonality(0)
    assert chord.chord_pitches == [0, 4, 7]

def test_chord_from_element():
    chord = Element(0) % Element(3).s.m
    assert chord.scale_pitches == [6, 8, 9, 11, 13, 14, 17]


def test_chord_notes_to_scale_pitches_with_octave():
    chord = (I % I.M.o(1)).o(1)
    pitch = chord.to_pitch(c0)

    assert pitch == 24



def test_patternize_melody():
    chord = ((I % I.M)(
        piano__0=s0.h.mp + s4.o(-1).mp,
        piano__3=s0.h.o(-1).p + s4.o(-2).p))
    expected_pattern = (
    (I % I.M)(
        v__0=x0.h.mp + bd1.mp))

    data, pattern = chord.patternize(melody=True)

    assert pattern == expected_pattern
    assert data['metadata']['melody'] == True
    assert data['voicing'] == [b0]
    assert data['metadata']['bar_duration'] == 3


def test_patternize_melody_with_octave_voicing():
    chord = ((I % VII.b.M)(
        piano__0=s0.h.mp + s4.o(-1).mp,
        piano__3=s0.h.o(-1).p + s4.o(-2).p))
    expected_pattern = (
    (I % I.M)(
        v__0=x0.h.mp + bd1.mp))

    data, pattern = chord.patternize(melody=True)

    assert pattern == expected_pattern
    assert data['metadata']['melody'] == True
    assert data['voicing'] == [b0.o(1)]
    assert data['metadata']['bar_duration'] == 3


def test_patternize_acc():
    chord = ((I % VII.b.M)(
        piano__0=s0.h.mp + s4.o(-1).mp,
        piano__3=s0.h.o(-1).p + s4.o(-2).p))
    expected_pattern = (
    (I % I.M)(
        v__0=x0.h.p + bd1.p))

    data, pattern = chord.patternize(melody=False, nb_excluded_instruments=1)

    assert pattern == expected_pattern
    assert data['metadata']['melody'] == False
    assert data['voicing'] == [b0]
    assert data['metadata']['bar_duration'] == 3





"""Run the network to annotate an unseen musical input (inference)."""

import os

import music21
import numpy as np
import pandas as pd
import re


from . import __version__
from .chord_vocabulary import frompcset, cosineSimilarity
from .cache import forceTonicization, getTonicizationScaleDegree
from .score_parser import parseScore, m21Parse
from .input_representations import available_representations as availableInputs
from .output_representations import (
    available_representations as availableOutputs,
)
from .utils import tensorflowGPUHack, disableGPU, padToSequenceLength
from tensorflow import keras

modelPath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'AugmentedNet.hdf5')

MODEL = None

def get_model():
    global MODEL
    print('LOADING MODEL')
    if MODEL is not None:
        return MODEL
    else:
        disableGPU()
        MODEL = keras.models.load_model(modelPath)
        return MODEL



inversions = {
    "triad": {
        0: "",
        1: "6",
        2: "64",
    },
    "seventh": {
        0: "7",
        1: "65",
        2: "43",
        3: "2",
    },
}


def formatChordLabel(cl):
    """Format the chord label for end-user presentation.

    Parameters
    ----------
    cl :
        

    Returns
    -------

    """
    # The only change I can think of: Cmaj -> C
    cl = cl.replace("maj", "") if cl.endswith("maj") else cl
    cl = cl.replace("-", "b")
    return cl


def formatRomanNumeral(rn, key):
    """Format the Roman numeral label for end-user presentation.

    Parameters
    ----------
    rn :
        
    key :
        

    Returns
    -------

    """
    # Something of "I" and "I" of something
    if rn == "I/I":
        rn = "I"
    return rn


def solveChordSegmentation(df):
    """

    Parameters
    ----------
    df :
        

    Returns
    -------

    """
    return df[df.HarmonicRhythm7 == 0].dropna()


def resolveRomanNumeralCosine(b, t, a, s, pcs, key, numerator, tonicizedKey):
    """

    Parameters
    ----------
    b :
        
    t :
        
    a :
        
    s :
        
    pcs :
        
    key :
        
    numerator :
        
    tonicizedKey :
        

    Returns
    -------

    """
    pcsetVector = np.zeros(12)
    chord = music21.chord.Chord(f"{b}2 {t}3 {a}4 {s}5")
    for n in chord.pitches:
        pcsetVector[n.pitchClass] += 1
    for pc in pcs:
        pcsetVector[pc] += 1
    chordNumerator = music21.roman.RomanNumeral(
        numerator.replace("Cad", "Cad64"), tonicizedKey
    ).pitchClasses
    for pc in chordNumerator:
        pcsetVector[pc] += 1
    smallestDistance = -2
    for pcs in frompcset:
        v2 = np.zeros(12)
        for p in pcs:
            v2[p] = 1
        similarity = cosineSimilarity(pcsetVector, v2)
        if similarity > smallestDistance:
            pcset = pcs
            smallestDistance = similarity
    if tonicizedKey not in frompcset[pcset]:
        # print("Forcing a tonicization")
        candidateKeys = list(frompcset[pcset].keys())
        # prioritize modal mixture
        tonicizedKey = forceTonicization(key, candidateKeys)
    rnfigure = frompcset[pcset][tonicizedKey]["rn"]
    chord = frompcset[pcset][tonicizedKey]["chord"]
    quality = frompcset[pcset][tonicizedKey]["quality"]
    chordtype = "seventh" if len(pcset) == 4 else "triad"
    # if you can't find the predicted bass
    # in the pcset, assume root position
    inv = chord.index(b) if b in chord else 0
    invfigure = inversions[chordtype][inv]
    if invfigure in ["65", "43", "2"]:
        rnfigure = rnfigure.replace("7", invfigure)
    elif invfigure in ["6", "64"]:
        rnfigure += invfigure
    rn = rnfigure
    if numerator == "Cad" and inv == 2:
        rn = "Cad64"
    if tonicizedKey != key:
        denominator = getTonicizationScaleDegree(key, tonicizedKey)
        rn = f"{rn}/{denominator}"
    chordLabel = f"{chord[0]}{quality}"
    if inv != 0:
        chordLabel += f"/{chord[inv]}"
    return rn, chordLabel


def generateRomanText(h, ts):
    """

    Parameters
    ----------
    h :
        
    ts :
        

    Returns
    -------

    """
    metadata = h.metadata
    metadata.composer = metadata.composer or "Unknown"
    metadata.title = metadata.title or "Unknown"
    composer = metadata.composer.split("\n")[0]
    title = metadata.title.split("\n")[0]
    rntxt = f"""\
Composer: {composer}
Title: {title}
Analyst: AugmentedNet v{__version__} - https://github.com/napulen/AugmentedNet
"""
    currentMeasure = -1
    for n in h.stream().flat.notes:
        if not n.lyric:
            continue
        rn = n.lyric.split()[0]
        key = ""
        measure = n.measureNumber
        beat = float(n.beat)
        if beat.is_integer():
            beat = int(beat)
        newts = ts.get((measure, beat), None)
        if newts:
            rntxt += f"\nTime Signature: {newts}\n"
        if ":" in rn:
            key, rn = rn.split(":")
        if measure != currentMeasure:
            rntxt += f"\nm{measure}"
            currentMeasure = measure
        if beat != 1:
            rntxt += f" b{round(beat, 3)}"
        if key:
            rntxt += f" {key.replace('-', 'b')}:"
        rntxt += f" {rn}"
    return rntxt


def predict(model, inputPath):
    """

    Parameters
    ----------
    model :
        
    inputPath :
        

    Returns
    -------

    """
    df = parseScore(inputPath)
    inputs = [l.name.rsplit("_")[1] for l in model.inputs]
    encodedInputs = [availableInputs[i](df) for i in inputs]
    outputLayers = [l.name.split("/")[0] for l in model.outputs]
    seqlen = model.inputs[0].shape[1]
    modelInputs = [
        padToSequenceLength(i.array, seqlen, value=-1) for i in encodedInputs
    ]
    predictions = model.predict(modelInputs)
    predictions = [p.reshape(1, -1, p.shape[2]) for p in predictions]
    dfdict = {}
    for outputRepr, pred in zip(outputLayers, predictions):
        predOnehot = np.argmax(pred[0], axis=1).reshape(-1, 1)
        decoded = availableOutputs[outputRepr].decode(predOnehot)
        dfdict[outputRepr] = decoded
    dfout = pd.DataFrame(dfdict)
    scoreLength = len(dfout.index)
    paddedIndex = np.full((scoreLength,), np.nan)
    paddedMeasure = np.full((scoreLength,), np.nan)
    paddedIndex[: len(df.index)] = df.index
    paddedMeasure[: len(df.s_measure)] = df.s_measure
    dfout["offset"] = paddedIndex
    dfout["measure"] = paddedMeasure
    chords = solveChordSegmentation(dfout)
    #s = music21.converter.parse(inputPath, forceSource=True)
    s = m21Parse(inputPath)
    ts = {
        (ts.measureNumber, float(ts.beat)): ts.ratioString
        for ts in s.flat.getElementsByClass("TimeSignature")
    }
    schord = s.chordify().flat.notesAndRests
    schord.metadata = s.metadata
    # remove all lyrics from score
    # for note in s.recurse().notes:
    #     note.lyrics = []
    prevkey = ""
    for analysis in chords.itertuples():
        notes = []
        for n in schord.getElementsByOffset(analysis.offset):
            if isinstance(n, music21.note.Note):
                notes.append((n, n.pitch.midi))
            elif isinstance(n, music21.chord.Chord) and not isinstance(
                n, music21.harmony.NoChord
            ):
                notes.append((n, n[0].pitch.midi))
        if not notes:
            continue
        bass = sorted(notes, key=lambda n: n[1])[0][0]
        thiskey = analysis.LocalKey38
        tonicizedKey = analysis.TonicizedKey38
        pcset = analysis.PitchClassSet121
        numerator = analysis.RomanNumeral31
        rn2, chordLabel = resolveRomanNumeralCosine(
            analysis.Bass35,
            analysis.Tenor35,
            analysis.Alto35,
            analysis.Soprano35,
            pcset,
            thiskey,
            numerator,
            tonicizedKey,
        )
        if thiskey != prevkey:
            rn2fig = f"{thiskey}:{rn2}"
            prevkey = thiskey
        else:
            rn2fig = rn2
        bass.addLyric(formatRomanNumeral(rn2fig, thiskey))
        bass.addLyric(formatChordLabel(chordLabel))
    rntxt = generateRomanText(schord, ts)
    filename, _ = inputPath.rsplit(".", 1)
    annotatedRomanText = f"{filename}_annotated.rntxt"
    with open(annotatedRomanText, "w") as fd:
        fd.write(rntxt)


def infer_chords(inputPath, useGpu=False):
    """

    Parameters
    ----------
    inputPath :
        
    useGpu :
         (Default value = False)

    Returns
    -------

    """

    if not os.path.isdir(inputPath):
        predict(get_model(), inputPath)
    for root, _, files in os.walk(inputPath):
        for f in files:
            name, ext = os.path.splitext(f)
            if ext not in [".mxl", ".xml", ".musicxml", ".krn"]:
                continue
            if "_annotated" in name:
                # do not recursively annotate an annotated_file
                continue
            filepath = os.path.join(root, f)
            predict(get_model(), inputPath=filepath)




import numpy as np
import pandas as pd
from musiclang.write.elements import ChordElement
import glob
import os
import joblib


class MelodyArrangerTrainer:

    def __init__(self, files_regex, output_directory):
        """

        Parameters
        ----------
        files_regex: str
            File regex that will be used for training
        output_directory: str
            Where the arrangement model will be saved
        """
        self.directory = files_regex
        self.output_directory = output_directory


    def train(self, nb_candidates=None):
        """
        Train the arranger model
        Parameters
        ----------
        nb_candidates: int
            Number of note type

        Returns
        -------

        """
        files = glob.glob(self.directory)
        self.texts = []
        for file in files:
            with open(file, 'r') as f:
                self.texts.append(f.read())
        states, priors, transitions, emissions = self.find_melodies(self.texts)

        # Save
        os.makedirs(self.output_directory, exist_ok=True)
        filepath_states = os.path.join(self.output_directory, 'states.pickle')
        filepath_priors = os.path.join(self.output_directory, 'priors.pickle')
        filepath_transitions = os.path.join(self.output_directory, 'transitions.pickle')
        filepath_emissions = os.path.join(self.output_directory, 'emissions.pickle')

        if nb_candidates is not None:
            # Find the most probable candidates
            keep_index = {mode: priors[mode].sort_values(ascending=False).index[:nb_candidates]
                          for mode in priors.keys()}
            priors = {mode: priors[mode].loc[keep_index[mode]]
                              for mode in priors.keys()}

            transitions = {mode: transitions[mode].loc[keep_index[mode], keep_index[mode]]
                                for mode in transitions.keys()}
            emissions = {mode: emissions[mode].loc[keep_index[mode]]
                              for mode in emissions.keys()}

            states = {mode: [s for s in states[mode] if s in keep_index[mode]] for mode in states.keys()}

        joblib.dump(states, filepath_states)
        joblib.dump(self._convert_to_dict(priors), filepath_priors)
        joblib.dump(self._convert_to_dict(transitions), filepath_transitions)
        joblib.dump(self._convert_to_dict(emissions), filepath_emissions)


    def _convert_to_dict(self, x):
        return {key: val.to_dict() for key, val in x.items()}


    def find_melodies(self, texts):
        sequences = []
        indexes = set()

        # First construct the sequences
        for text in texts:
            lines = text.split('\n')
            lines = [l for l in lines if l != '' and l[0] == 'm']
            for line in lines:
                for el in line.split(' '):
                    if el == '' or el[0] == 'm':
                        pass
                    elif el.startswith('b') and el[1].isdigit():
                        pass
                    elif ':' in el and not ('|' in el):
                        new_tonality = el.replace(':', '')
                        if new_tonality != tonality:
                            tonality = new_tonality
                            sequences.append([])
                    else:
                        sequences[-1].append(el)
                        indexes.add(el)

        # Create the proper indexes
        indexes_list = {mode: list(value) for mode, value in indexes.items()}
        indexes = {mode: {val: i for i, val in enumerate(value)} for mode, value in indexes_list.items()}
        reversed_indexes = {mode: {val: key2 for key2, val in value.items()} for mode, value in indexes.items()}

        # Construct the pitches candidates
        pitches_candidates = {'m': {}, 'M': {}}
        for mode, tone in [('m', 'c'), ('M', 'C')]:
            for pitch in range(12):
                pitches_candidates[mode][pitch] = set()
                for chord in indexes_list[mode]:
                    try:
                        if ChordElement(chord).has_pitch(pitch, tone):
                            pitches_candidates[mode][pitch].add(chord)
                    except:
                        pass

        # Construct Prior proba of chords and transition probabilities
        priors = {mode: np.zeros(len(index), dtype=np.float32) for mode, index in indexes.items()}
        transitions = {mode: np.zeros((len(index), len(index)), dtype=np.float32) for mode, index in indexes.items()}
        for mode in ['m', 'M']:
            for sequence in sequences[mode]:
                transitions_seq = list(zip(sequence, sequence[1:]))
                for chord in sequence:
                    priors[mode][indexes[mode][chord]] += 1
                for chord1, chord2 in transitions_seq:
                    transitions[mode][indexes[mode][chord1], indexes[mode][chord2]] += 1

        # Normalize probabilities
        transitions = {mode: np.nan_to_num(val / val.sum(axis=1)[:, np.newaxis]) for mode, val in transitions.items()}
        priors = {mode: val / val.sum() for mode, val in priors.items()}
        transitions = {mode: pd.DataFrame(transitions[mode],
                                          index=indexes_list[mode],
                                          columns=indexes_list[mode]) for mode in ['m', 'M']}
        priors = {mode: pd.Series(priors[mode],
                                  index=indexes_list[mode]) for mode in ['m', 'M']}

        # Create emission matrix
        emissions = {mode: pd.DataFrame(np.zeros((len(index), 12), dtype=np.float32),
                                        index=indexes_list[mode],
                                        columns=list(range(12)))
                     for mode, index in indexes.items()}
        for mode in ['m', 'M']:
            for i in range(12):
                emission = emissions[mode]
                emission.loc[list(pitches_candidates[mode][i]), i] = 1

        states = indexes_list
        return states, priors, transitions, emissions






import numpy as np


class PitchDynamizer:
    """
    Create the dynamics with the average pitch of each chord
    If ascending is true : A higher average pitch means an higher dynamic for this dynamizer
    if ascending is false, the contrary
    """

    def __init__(self, minimum=30, maximum=110, ascending=True):
        self.minimum = minimum
        self.maximum = maximum
        self.ascending = ascending

    def __call__(self, score, **kwargs):

        # Calculate mean pitch per chords

        pitches = []
        for chord in score.chords:
            pitches_ = [chord.to_pitch(n) for part in chord.parts for n in chord.score[part].notes if n.is_note]
            pitches.append(np.mean(pitches_))

        # min max
        minimum = self.minimum
        maximum = self.maximum

        pitches = [(p - np.min(pitches)) / (np.max(pitches) - np.min(pitches)) for p in pitches]
        if not self.ascending:
            amps = [(p * minimum) + (1 - p) * maximum for p in pitches]
        else:
            amps = [(p * maximum) + (1 - p) * minimum for p in pitches]

        for i in range(5):
            amps = [(a + b)/2 for a, b in zip(amps, [amps[0]] + amps[:-1])]

        new_score = None
        for chord, amp in zip(score.chords, amps):
            new_score += chord.set_amp(amp)

        return new_score


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

from .constants import *
LIMIT_DENOM = int(1e4)

class Note:
    """
    Represents a note in MusicLang.

    **The specificity of MusicLang is that notes are always represented relatively to a chord inside a tonality.**

    Usually you won't be instantiating notes yourself but use the builtin ``musiclang.write.library`` which already
    defines common symbols of musiclang. See the examples to get familiar with the notation library.


    .. warning:: Notes are 0-indexed relative to a scale, so for example ``s0`` is the first note of the scale.

    Note Value
    ----------

    There are different kinds of notes

    - Scale notes : ``s0, s1, s2, s3, s4, s5, s6`` : relative to 7 sounds scale of the chord inside a tonality
    - Chromatic notes : ``h0, h1, h2, h3, h4, h5, h6, h7, h8, h9, h10, h11`` : \
    relative to the 12 sounds of the chromatic scale associated with the chord/tonality
    - Relative up scale note, ``su0, su1, su2, su3, su4, su5, su6`` : Up relatively to the previous note of a melody
    - Relative down scale note, ``sd0, sd1, sd2, sd3, sd4, sd5, sd6`` : Down relatively to the previous note of a melody
    - Relative up chromatic note, ``hu0, hu1, hu2,  hu3, hu4, hu5, hu6, hu7, hu8, hu9, hu10, hu11`` : \
    Up relatively to the previous note of a melody
    - Relative down chromatic note, ``hd0, hd1, hd2,  hu3, hu4, hu5, hu6, hu7, hu8, hu9, hu10, hu11`` : \
     Down relatively to the previous note of a melody

    Octaves
    -------

    You can specify an octave using the :func:`~o` method of a note
    For example : ``s0.o(1)`` is ``s0`` up one octave, ``s0.o(-1)`` is ``s0`` down one octave

    Rhythm
    -------

    You can spcify a rhythm to a note using properties

    - ``h=half, w=whole, q=quarter, e=eight, s=sixteenth, t=thirty-seconds``
    - You can use n-uplet ``(3, 5, 7)`` : for example s0.e3 is s0 with a duration of a triolet etc ...
    - You can use dots with ``d`` : For example s0.qd has a dotter quarter duration. You can use double dots
    - You can use the :func:`~Note.augment` method if you want a custom duration that can't be notated easily. \
     For example ``s0.augment(8)`` for a duration of two whole notes)

    .. note:: You must use the builtin python `fractions.Fraction` object to create \
    duration otherwise you will get rounding errors when exporting to midi.

    Dynamics
    --------

    You can add dynamics to a note using ``ppp, pp, p, mf, f, ff, fff`` properties of the note
    Example : s0.fff is a note in triple forte

    Silences
    --------

    You can specify a silence using the ``r`` notation in the library, otherwise use the Silence class

    Continuation
    ------------
    You can specify a note continuation using the ``l`` notation in the library, otherwise use the Continuation class

    Mode
    ----

    You can force a mode on a note that bypass the mode of the chord scale.

    Accident
    ---------

    You can force an accident on a note choosing between : ['min', 'maj', 'dim', 'aug', 'natural']
    It is used to force a specific interval that is robust to transposition, but keeping a diatonic value.
    It is advised to be used in replacement of h-type notes because it is in general more compatible with transformations.

    Tempo
    -----

    You can set a new tempo when this note is played, it will be played


    Pedal
    -----

    You can set the pedal on or off when this note is played, it will be applied to the whole instrument, not a single part

    Examples
    --------
    >>> from musiclang.library import s0, s1, s2, I

    You can learn what pitch is associated to a note in the context of a chord :

    >>> chord = (I % I.M) # C major chord (first degree of first major tonality relative to C)
    >>> note = s0
    >>> chord.to_pitch(s0)
    0

    >>> chord.to_pitch(s1)
    2

    >>> chord.to_pitch(s2)
    4

    0 is the note C5 in musiclang, so 2 is D5

    >>> (I % I.m).to_pitch(s2)
    3

    In minor s2 is eb (=3)
    """
    DEFAULT_AMP = 66

    def __init__(self, type, val, octave, duration, mode=None, accident=None, amp=DEFAULT_AMP, tags=None, pedal=None, tempo=None):
        self.type = type
        self.val = val
        self.octave = octave
        self.duration = frac(duration).limit_denominator(LIMIT_DENOM)
        self.amp = amp
        self.accident = accident
        self.mode = mode
        self.properties = self.init_properties()
        self.tags = tags if tags is not None else set()
        self.pedal = pedal
        self.tempo = tempo

    def has_tag(self, tag):
        """
        Check if the tag exists for this note
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        note: Note
        """
        return tag in self.tags

    def add_tag(self, tag):
        """
        Add a tag to this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        note: Note
        """
        cp = self.copy()
        cp.tags.add(tag)
        return cp

    def add_tags(self, tags):
        """
        Add several tags to the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        note: Note

        """
        cp = self.copy()
        cp.tags = cp.tags.union(set(tags))
        return cp

    @property
    def interpolate(self):
        return self.add_tag('interpolate')

    @property
    def accent(self):
        return self.add_tag('accent')

    @property
    def mordant(self):
        return self.add_tag('mordant')

    @property
    def chroma_mordant(self):
        return self.add_tag('chroma_mordant')

    @property
    def inv_chroma_mordant(self):
        return self.add_tag('inv_chroma_mordant')


    @property
    def inv_mordant(self):
        return self.add_tag('inv_mordant')

    @property
    def grupetto(self):
        return self.add_tag('grupetto')

    @property
    def inv_grupetto(self):
        return self.add_tag('inv_grupetto')

    @property
    def chroma_grupetto(self):
        return self.add_tag('chroma_grupetto')

    @property
    def inv_chroma_grupetto(self):
        return self.add_tag('inv_chroma_grupetto')

    @property
    def roll(self):
        return self.add_tag('roll')

    @property
    def roll_fast(self):
        return self.add_tag('roll_fast')

    @property
    def suspension_prev(self):
        return self.add_tag('suspension_prev')

    @property
    def suspension_prev_repeat(self):
        return self.add_tag('suspension_prev_repeat')

    @property
    def retarded(self):
        return self.add_tag('retarded')

    def realize_tags(self, last_note=None, next_note=None):
        from .ornementation import realize_tags

        return realize_tags(self, last_note, next_note)




    def set_tempo(self, tempo):
        new_note = self.copy()
        new_note.tempo = tempo
        return new_note

    @property
    def pedal_on(self):
        new_note = self.copy()
        new_note.pedal = True
        return new_note

    @property
    def pedal_off(self):
        new_note = self.copy()
        new_note.pedal = False
        return new_note

    def remove_effects(self):
        """
        Remove pedals and tempo change
        Returns
        -------
        """
        new_note = self.copy()
        new_note.tempo = None
        new_note.pedal = None
        return new_note

    def remove_tempo(self):
        """
        Remove tempo change
        Returns
        -------
        """
        new_note = self.copy()
        new_note.tempo = None
        return new_note

    def remove_pedal(self):
        """
        Remove pedal
        Returns
        -------
        """
        new_note = self.copy()
        new_note.pedal = None
        return new_note

    def remove_tags(self, tags):
        """
        Remove several tags from the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]

        Returns
        -------
        note: Note


        """
        cp = self.copy()
        cp.tags = cp.tags - set(tags)
        return cp

    def clear_note_tags(self):
        return self.clear_tags()

    def remove_tag(self, tag):
        """
        Remove a tag from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        note: Note
        """
        cp = self.copy()
        cp.tags.remove(tag)
        return cp

    def clear_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        note: Note
        """
        cp = self.copy()
        cp.tags = set()
        return cp

    def __iter__(self):
        return [self].__iter__()

    def real_chord(self, chord):
        """

        If the note has a mode

        Parameters
        ----------
        chord :


        Returns
        -------

        """
        new_chord = chord.copy()
        if self.mode is None:
            return new_chord
        else:
            new_chord = new_chord.change_mode(self.mode)
            return new_chord

    def decompose_duration(self):
        """
        Decompose a note into note and continuations with existing figures.
        It tries to avoid that the string representation of a note contains any "augment" calls
        """

        def _recurse(note):
            """
            Recursive sub method to decompose a note into smaller values note and continuations
            Parameters
            ----------
            note : musiclang.Note

            Returns
            -------
            Note or Melody

            Examples
            --------

            >>> from musiclang.library import s0
            >>> from fractions import Fraction
            >>> s0.augment(Fraction(11, 8)).decompose_duration()
            s0 + l.s + l.t


            """
            from musiclang.write.constants import DURATION_TO_STR
            from fractions import Fraction as frac

            if note.duration in DURATION_TO_STR:
                return note

            duration = frac(note.duration)
            candidates = [frac(d) for d in DURATION_TO_STR.keys() if d != 0]
            candidates = [c for c in candidates if (duration / c).denominator == 1 and c < duration]
            if len(candidates) == 0:
                return note
            chosen_candidate = max(candidates)
            base_note = note.augment(chosen_candidate / duration)
            new_note = note.augment((duration - base_note.duration) / duration)
            return base_note + _recurse(Continuation(new_note.duration))

        result = _recurse(self)
        # Reverse the melody
        if len(result.notes) > 1:
            result = result[::-1]
            dur = result.notes[0].duration
            result.notes[0] = result.notes[-1].copy().augment(dur / result.notes[-1].duration)
            result.notes[-1] = Continuation(result.notes[-1].duration)
        return result

    def init_properties(self):
        """ """
        from .properties.note_properties import NoteProperties
        return NoteProperties(self)

    def copy(self):
        """ """
        return Note(self.type, self.val, self.octave, self.duration, mode=self.mode, accident=self.accident,
                    amp=self.amp, tags=set(self.tags),
                    tempo=self.tempo,
                    pedal=self.pedal
                    )

    def repeated_notes_to_legato(self):
        return self.to_melody().repeated_notes_to_legato()

    def set_amp(self, amp):
        new_note = self.copy()
        if isinstance(amp, float):
            amp = int(amp)
        if isinstance(amp, int):
            new_note.amp = amp
        elif isinstance(amp, str):
            return self.__getattr__(amp)

        return new_note

    def set_val(self, val):
        note = self.copy()
        note.val = val
        return note


    def project_on_rhythm(self, rhythm, **kwargs):
        return self.to_melody().project_on_rhythm(rhythm, **kwargs)

    def to_scale_note(self, chord):
        if not self.is_note:
            return self.copy()
        return chord.parse(chord.to_pitch(self)).set_duration(self.duration).set_amp(self.amp).add_tags(self.tags)

    def to_scale_notes(self, chord):
        from musiclang import Melody
        return Melody(self.to_scale_note(chord))

    def set_duration(self, value):
        """
        Set the duration of a note
        Parameters
        ----------
        value :


        Returns
        -------

        """
        if isinstance(value, float):
            value = frac(value).limit_denominator(LIMIT_DENOM)
        elif isinstance(value, int):
            value = frac(value, 1)
        result = self.copy()
        result.duration = value
        result.duration = result.duration.limit_denominator(LIMIT_DENOM)
        return result

    def augment(self, value):
        """
        Returns a copy with augmented duration (multiply current duration by the value)

        Parameters
        ----------
        value : fractions.Fraction
                Fraction on which to multiply the current duration


        Returns
        -------
        note: Note
              New note with increased duration

        Examples
        --------

        >>> from musiclang.library import s0
        >>> from fractions import Fraction
        >>> s0.augment(Fraction(8, 7))
        s0.augment(frac(8, 7))

        """
        if isinstance(value, float):
            value = frac(value).limit_denominator(LIMIT_DENOM)
        if isinstance(value, int):
            value = frac(value, 1)
        result = self.copy()
        result.duration *= value
        result.duration = result.duration.limit_denominator(LIMIT_DENOM)
        return result

    def change_type(self, new_type):
        """

        Parameters
        ----------
        new_type :


        Returns
        -------

        """
        c = self.copy()
        c.type = new_type
        return c

    def add_interval(self, note):
        """

        Parameters
        ----------
        note :


        Returns
        -------

        """
        if note.type[1] not in ['u', 'd']:
            raise Exception('note argument must be a relative type (*u, *d)')
        val = note.val if note.type[1] == 'u' else -note.val
        oct = note.octave
        return self.add_value(val, oct)

    def add_value(self, val, octave):
        """

        Parameters
        ----------
        val :

        octave :


        Returns
        -------

        """
        if self.type[0] in ['s', 'h', 'c']:
            if self.type[0] == 's':
                mod = 7

            elif self.type[0] == 'h':
                mod = 12

            elif self.type[0] == 'c':
                mod = 3

            new_note = self.copy()
            new_note.val += val
            new_note.octave += octave
            new_note.octave += new_note.val // mod
            new_note.val = new_note.val % mod
        else:
            return self.copy()

        return new_note

    def oabs(self, octave):
        """

        Parameters
        ----------
        octave :


        Returns
        -------

        """
        note = self.copy()
        note.octave += octave
        return note

    def o(self, octave):
        """
        Relative octave (change if definite absolute do not otherwise)

        Parameters
        ----------
        octave : int
            Number of octave (positive or negative) to transpose the current note


        Returns
        -------
        note: Note
            It will return a new note transposed

        """
        if self.type in ['s', 'h', 'c', 'b', 'a', 'x']:
            return self.oabs(octave)
        return self.copy()

    def __or__(self, other):
        return self.to_melody() | other


    def duration_to_str(self):
        """ """
        return DURATION_TO_STR[self.duration]


    def to_drum(self):
        note = self.copy()
        if not note.is_silence and not note.is_continuation:
            note.type = 'd'
        return note

    @property
    def base_octave(self):
        return 5

    @property
    def is_drum(self):
        return self.type == 'd'

    def replace(self, to_replace, new_note, octave=False, amp=False, add_octave=True):
        """
        Replace a note by another note

        Parameters
        ----------
        to_replace :

        new_note :


        Returns
        -------

        """

        note = self
        to_add = note.copy()
        test = note.val == to_replace.val and note.type == to_replace.type
        test &= (note.octave == to_replace.octave) or not octave
        if test:
            to_add.type = new_note.type
            to_add.val = new_note.val
            to_add.octave = new_note.octave
            if not octave:
                to_add.octave += note.octave
            to_add.mode = new_note.mode
            to_add.accident = new_note.accident
            if amp:
                to_add.amp = note.amp

        return to_add

    def to_code(self):
        """
        Represent a note as valid python code using standard musiclang library

        Parameters
        ----------

        Returns
        -------

        A string representation of the note valid using standard musiclang library

        """

        if self.is_note:
            result = f"{self.type}{self.val}"
        elif self.is_drum:
            result = f'd{self.val}'
            if self.octave != 0:
                result += f".oabs({self.octave})"
        elif self.type == 'x':
            result = f"{self.type}{self.val}"
        else:
            result = f"{self.type}"

        if self.duration != Q:
            if self.duration in DURATION_TO_STR:
                duration = DURATION_TO_STR[self.duration]
                result += f".{duration}"
            else:
                if isinstance(self.duration, int):
                    result += f".augment({self.duration}))"
                else:
                    result += f".augment(frac({self.duration.numerator}, {self.duration.denominator}))"

        if self.octave != 0 and self.is_note:
            if not self.is_relative:
                result += f".o({self.octave})"
            else:
                result += f".oabs({self.octave})"

        if self.mode is not None and self.is_note:
            result += f".{self.mode}"
        if self.accident is not None and self.is_note:
            result += f".{self.accident}"
        if self.is_note or self.type == "x":
            amp_figure = self.amp_figure
            if amp_figure != 'mf':
                result += f".{self.amp_figure}"
        if len(self.tags) > 0:
            result += f".add_tags({self.tags})"

        return result


    def remove_accidents(self):
        new_note = self.copy()
        new_note.accident = None
        return new_note

    @property
    def has_accident(self):
        return self.accident is not None

    def to_sequence(self, chord, inst):
        """
        Helper function to generate an array of some specific property from a note
        .. warning:: Only used internally to generate a dataframe from a Score

        Transform in a list of [(start_time, end_time, pitch, self)]

        Parameters
        ----------
        chord : Chord
            Chord on which the note will be played


        inst : str
            Instrument on which the note will be played

        Returns
        -------
        A List of list with one element

        Examples
        --------

        >>> from musiclang.library import *
        >>> s0.to_sequence(I % I.M, 'piano__0')
        [[0,1,0,(I % I.M)(),'piano__0',s0]]
        """

        from .melody import Melody
        return Melody([self]).to_sequence(chord, inst)

    def to_melody(self):
        from .melody import Melody
        return Melody([self], tags=self.tags)

    def repr_mode(self):
        """ """
        if self.mode is not None:
            return f".{self.mode}"

        return ""



    def __repr__(self):
        return self.to_code()

    def __len__(self):
        return 1

    def __add__(self, other):
        """
        Create a melody from the note and other note or melody
        Parameters
        ----------
        other : Note or Melody

        Returns
        -------

        A melody with the concatenated notes

        """
        from .melody import Melody
        if other is None:
            return self.copy()
        if isinstance(other, Note):
            return Melody([self, other])
        if isinstance(other, Melody):
            return Melody([self] + other.notes)
        else:
            raise Exception('Cannot add if not instance of Note or melody')

    def __and__(self, n):
        """
        Transpose the value of the note by n

        Parameters
        ----------

        n : int
            Value to transpose the note

        Examples
        --------

        >>> from musiclang.library import *
        >>> s0 & 2
        s2
        """
        if self.type in ['s', 'h']:
            return self.add_value(n, 0)
        elif self.type in ['h']:
            return self.add_value((12 * n) // 7, 0)
        else:
            return self.copy()

    def __matmul__(self, other):
        """
        Apply a function to this note
        The function should have the signature ``f(note: Note) -> Note | Melody``

        Parameters
        ----------
        other : function
                Function to apply to this note

        Returns
        -------
        note: Note
              Note with applied function

        Examples
        --------

        >>> from musiclang.library import *
        >>> f = lambda note: note.fff
        >>> s0 @ f
        s0.fff


        """
        # Apply a function to each note
        return other(self.copy())

    def convert_to_drum_note(self, chord):
        if self.type.startswith('d') or self.is_silence or self.is_continuation:
            return self.copy()
        new_note = self.copy()
        new_note.type = 'd'
        pitch = chord.to_pitch(self)
        new_note.val = pitch % 12
        new_note.octave = pitch // 12
        return new_note

    def to_standard_note(self, chord):
        if self.type in ['b', 'c']:
            if self.is_bass_note:
                candidates = chord.extension_notes
            elif self.is_chord_note:
                candidates = chord.chord_notes
            else:
                raise Exception('Not existing type')
            val = self.val
            octave = self.octave
            new_note = candidates[val % len(candidates)].o((val // len(candidates)) + octave)
            new_note = new_note.set_duration(self.duration)
            new_note.amp = self.amp
            return new_note

        elif self.type in ['a']:
            new_note = self.copy()
            base_note = chord.parse(chord.to_pitch(self))
            new_note.type = base_note.type
            new_note.val = base_note.val
            new_note.octave = base_note.octave
            return new_note
        else:
            return self.add_tags(self.tags)

    def to_absolute_note(self, chord, last_pitch=None):
        if not self.is_note:
            return self.copy()
        pitch = chord.to_pitch(self, last_pitch=last_pitch)
        new_note = self.copy()
        new_note.type = 'a'
        new_note.val = pitch % 12
        new_note.octave = pitch // 12

        return new_note

    def as_key(self, octave=False, duration=False, amp=False):
        oct = self.octave if octave else 0
        duration = self.duration if duration else 1
        amp = self.amp if amp else 66
        return Note(self.type, self.val, oct, duration, mode=self.mode, amp=amp, accident=self.accident)


    def to_extension_note(self, chord):
        candidates = chord.extension_notes
        candidates_without_octave = [c.o(-c.octave) for c in candidates]
        try:
            idx = candidates_without_octave.index(self.as_key())
            note = candidates[idx]
            to_return = self.copy()
            to_return.type = 'b'
            to_return.val = idx
            to_return.octave -= note.octave
            return to_return
        except:
            return self.copy()

    def to_chord_note(self, chord):
        candidates = chord.chord_notes
        candidates_without_octave = [c.o(-c.octave) for c in candidates]
        try:
            idx = candidates_without_octave.index(self.as_key())
            note = candidates[idx]
            to_return = self.copy()
            to_return.type = 'c'
            to_return.val = idx
            to_return.octave -= note.octave
            return to_return
        except:
            return self.copy()


    def apply_pattern(self, *voicing):
        return self.to_melody().apply_pattern(*voicing)


    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        """
        Check if two notes are equals. To be equals note should have

        - Same type
        - Same value
        - Same duration
        - Same octave
        - Same mode

        Parameters
        ----------
        other : Other note to compare

        Returns
        -------
        is_equal: bool

        Examples
        --------

        >>> from musiclang.library import *
        >>> s0 == s1 & -1
        """
        if not isinstance(other, Note):
            return False
        else:
            return (self.type == other.type) and (self.val == other.val) and (self.duration == other.duration) and \
                   (self.octave == other.octave) and (self.mode == other.mode)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    def __getattr__(self, item):
        try:
            note = self.copy()
            note.duration *= STR_TO_DURATION[item]
            return note
        except KeyError:
            if hasattr(self.properties, item):
                return getattr(self.properties, item)
            else:
                raise AttributeError('Not existing properties of attribute {}'.format(item))

    def __mul__(self, other):
        """
        If other is Integer, repeat the note other times
        """
        from .melody import Melody
        if isinstance(other, int):
            return Melody([self.copy() for i in range(other)])
        else:
            raise Exception('Cannot multiply Note and ' + str(type(other)))

    def __radd__(self, other):
        from .melody import Melody
        if other is None:
            return Melody([self.copy()])
        else:
            raise Exception('Cannot add')


### DERIVED CLASSES OF NOTE

class Silence(Note):
    """
    Represents a silence in MusicLang.

    Only takes a duration as a parameter
    """

    def __init__(self, duration, tags=None, tempo=None, pedal=None):
        super().__init__("r", 0, 0, duration, tags=tags, tempo=tempo, pedal=pedal)

    def copy(self):
        """ """
        return Silence(self.duration, tempo=self.tempo, pedal=self.pedal, tags=set(self.tags))


class Continuation(Note):
    """
    Represents a continuation in MusicLang.

    Only takes a duration as a parameter
    """

    def __init__(self, duration, tags=None, tempo=None, pedal=None):
        super().__init__("l", 0, 0, duration, tempo=tempo, pedal=pedal, tags=tags)

    def copy(self):
        """ """
        return Continuation(self.duration, pedal=self.pedal, tags=set(self.tags))



from musiclang.transform import Mask
from musiclang.library import *


def test_mask_chord():
    """
    Mask.Chord() returns false if the object is not a chord
    Otherwise returns true
    """
    mask = Mask.Chord()
    assert not mask((I % I.M))
    assert mask(s0)


def test_mask_melody():
    """
    Mask.Melody() returns false if the object is not a chord
    Otherwise returns true
    """
    mask = Mask.Melody()
    assert not mask(s0 + s1)
    assert mask(s0)
    assert mask(I % I.M)


def test_mask_note():
    """
    Mask.Note() returns false if the object is not a chord
    Otherwise returns true
    """
    mask = Mask.Note()
    assert not mask(s0)
    assert mask((I % I.M))


def test_mask_score():
    """
    Mask.Score() returns false if the object is not a chord
    Otherwise returns true
    """
    mask = Mask.Score()
    assert not mask((I % I.M)(piano=[s0, s2, s4]) + (I % I.M)(piano=[s0, s2, s4]))
    assert mask(s0)


def test_mask_gt_true():

    mask = Mask.Chord() > Mask.Has('ok')
    score = (I % I.M)().add_tag('ok')
    assert mask(score)


def test_mask_gt_false():

    mask = Mask.Chord() > Mask.Has('nok')
    score = (I % I.M)().add_tag('ok')
    assert not mask(score)

def test_mask_and_true():

    mask = Mask.Chord() > (Mask.Has('ok1') & Mask.Has('ok2'))
    score = (I % I.M)().add_tags(['ok1', 'ok2'])
    assert mask(score)

def test_mask_and_false():

    mask = Mask.Chord() > (Mask.Has('ok1') & Mask.Has('ok2'))
    score = (I % I.M)().add_tags(['ok1'])
    assert not mask(score)

def test_mask_or_true():

    mask = Mask.Chord() > (Mask.Has('ok1') | Mask.Has('ok2'))
    score = (I % I.M)().add_tags(['ok1'])
    assert mask(score)

def test_mask_or_false():

    mask = Mask.Chord() > (Mask.Has('ok1') | Mask.Has('ok2'))
    score = (I % I.M)().add_tags(['ok3'])
    assert not mask(score)

def test_mask_hasatleast_true():

    mask = Mask.Chord() > (Mask.HasAtLeast(['ok1', 'ok2' ]))
    score = (I % I.M)().add_tags(['ok1'])
    assert mask(score)

def test_mask_hasatleast_false():

    mask = Mask.Chord() > (Mask.HasAtLeast(['ok1', 'ok2' ]))
    score = (I % I.M)().add_tags(['ok3'])
    assert not mask(score)

def test_not_mask():

    mask = (Mask.Melody() > Mask.Has('ok'))

    melo = (s0 + s1)

    assert not mask(melo)
    assert (~mask)(melo)

def test_bool_mask():
    mask = (Mask.Melody() > Mask.Bool(True))
    melo = (s0 + s1)
    assert mask(melo)

def test_eval_mask():

    mask = Mask.Chord() > (Mask.HasAtLeast(['ok1', 'ok2']))

    repr = str(Mask.eval(str(mask)))


from musiclang.library import *

import pytest


def test_normalization_extension():
    chord = (I % I.M)['6{-5}[m3](sus2)[add6]']
    assert chord.extension == '6(sus2)[add6][m3]{-5}'


def test_without_additions_or_replacements():
    chord = (I % I.M)['6']
    assert chord.extension_notes == [s2, s4, s0.o(1)]

def test_unexisting_extension_lead_to_error():
    with pytest.raises(Exception):
        chord = (I % I.M)['22zrahj']

def test_replacement_without_candidate_lead_to_addition():
    chord = (I % I.M)['[#11]']
    assert chord.extension_notes == [s0, s2, s4, h6.o(1)]


def test_sus2_replacement():

    chord = (I % I.M)['(sus2)']
    assert chord.extension_notes == [s0, s1, s4]


def test_remove_fifth():
    chord = (I % I.M)['{-5}']
    assert chord.extension_notes == [s0, s2]


def test_remove_eleventh():
    chord = (I % I.M)['13{-11}[add4]']
    assert chord.extension == '13[add4]{-11}'
    assert chord.extension_notes == [s0, s2, s3, s4, s6, s1.o(1), s5.o(1)]


def test_add2():
    chord = (I % I.M)['[add2]']
    assert chord.extension_notes == [s0, s1, s2, s4]

def test_very_complicated_chord():
    chord = (I % I.M)['9[add2][#11](m3)']
    assert chord.extension_notes == [s0, s1, h3, s4, s6, s1.o(1), h6.o(1)]


def test_very_complicated_chord2():
    chord = (I % I.M)['2[add4][M6](M3)(+)']
    assert chord.extension_notes == [s6.o(-1), s0, h4, s3, h8, h9]

def test_replacement_become_addition():
    chord = (I % I.M)['(M7)']
    assert chord.extension_notes == [s0, s2, s4, h11]


def test_replace_b9():
    chord = (I % I.M)['9(m9)']
    assert chord.extension_notes == [s0, s2, s4, s6, h1.o(1)]


def test_addition_and_replacement():
    chord = (I % I.M)['(sus2)[add6]']
    assert chord.extension_notes == [s0, s1, s4, s5]



def test_addition_and_inversion():
    chord = (I % I.M)['6(sus2)']
    assert chord.extension_notes == [s1, s4, s0.o(1)]



def test_addition_and_inversion_with_octaves():
    chord = (I % I.M)['2(sus2)']
    assert chord.extension_notes == [s6.o(-1), s0, s1, s4]


def test_to_voicing():
    chord = (I % I.M)['(sus2)']
    chord = chord.to_voicing(nb_voices=3, instruments=['cello', 'piano', 'flute'])
    assert chord == (I % I.M)['(sus2)'](cello__0=s0, piano__0=s1, flute__0=s4)



def test_score_to_voicing():
    score = (I % I.M)['(sus2)'] + (V % I.M)['[add9](+)']
    score = score.to_voicing(nb_voices=4, instruments=['cello', 'piano', 'flute', 'piccolo'])
    ch1 = (I % I.M)['(sus2)'](cello__0=s0, piano__0=s1, flute__0=s4, piccolo__0=s0.o(1))
    ch2 = (V % I.M)['[add9](+)'](cello__0=s0, piano__0=s2, flute__0=h8, piccolo__0=s1.o(1))
    assert score == ch1 + ch2

def test_score_to_voicing_with_duration():
    score = (I % I.M)['(sus2)'].w + (V % I.M)['[add9](+)'].augment(frac(6, 1))
    score = score.to_voicing(nb_voices=4, instruments=['cello', 'piano', 'flute', 'piccolo'])
    ch1 = (I % I.M)['(sus2)'](cello__0=s0.w, piano__0=s1.w, flute__0=s4.w, piccolo__0=s0.o(1).w)
    ch2 = (V % I.M)['[add9](+)'](cello__0=s0.augment(frac(6, 1)), piano__0=s2.augment(frac(6, 1)), flute__0=h8.augment(frac(6, 1)), piccolo__0=s1.o(1).augment(frac(6, 1)))
    assert score == ch1 + ch2


from musiclang.write.library import *



def test_score_equals():
    score1 = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    score2 = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    assert score1 == score2

def test_score_not_equals():
    score1 = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    score2 = I(piano__0=s0 + s1 + s1, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    assert score1 != score2

def test_score_equals_invert_order():
    score1 = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    score2 = I(violin__0=s0 + s1 + s4, piano__0=s0 + s1 + s2) + II(piano__0=s1 + s2)
    assert score1 == score2

def test_get_score_between():
    score = I(piano__0=s0 + s1 + s2) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=1, end=4)
    expected_result = I(piano__0=s1 + s2) + II(piano__0=s1)
    assert subscore == expected_result


def test_complex_simple_score_between():
    score = I(piano__0=s0 + s1 + s2) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=0, end=6)
    expected_result = score
    assert subscore == expected_result

def test_complex_out_of_bound_score_between():
    score = I(piano__0=s0 + s1 + s2) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=0, end=7)
    expected_result = score
    assert subscore == expected_result

def test_complex_out_of_bound_and_complex_score_between():
    score = I(piano__0=s0 + s1 + s2) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=1, end=7)
    expected_result =  I(piano__0=s1 + s2) + II(piano__0=s1 + s2)
    assert subscore == expected_result


def test_multi_instrument_complex_score_between():
    score = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=1, end=4)
    expected_result =  I(piano__0=s1 + s2, violin__0=s1 + s4) + II(piano__0=s1)
    assert subscore == expected_result

def test_multi_instrument_2_score_between():
    score = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=0, end=3)
    expected_result =  I(piano__0=s0 +s1 + s2, violin__0=s0 + s1 + s4)
    assert subscore == expected_result

def test_multi_instrument_3_score_between():
    score = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=3, end=6)
    expected_result =  II(piano__0=s1 + s2)
    assert subscore == expected_result


def test_multi_instrument_1_2_quarters_score_between():
    score = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)
    subscore = score.get_score_between(start=1, end=2)
    expected_result =  I(piano__0=s1, violin__0=s1)
    assert subscore == expected_result

def test_octaver():

    score = I(piano__0=s0 + s1 + s2, violin__0=s0 + s1 + s4) + II(piano__0=s1 + s2)

    score2 = score.octaver(piano__0=-1, violin__0=1)
    expected = I(piano__0=(s0 + s1 + s2).o(-1), violin__0=(s0 + s1 + s4).o(1)) + II(piano__0=(s1 + s2).o(-1))
    assert score2 == expected

def test_normalize_instruments():

    score = (I % I.M)(piano__0=s0.h) + (V % V.M)(piano__0=s1 + s2, violin__1=s3 + s1)
    expected_score = (I % I.M)(piano__0=s0.h, violin__1=r.h) + (V % V.M)(piano__0=s1 + s2, violin__1=s3 + s1)
    assert score.normalize_instruments() == expected_score

def test_remove_silenced_instruments():

    score = (I % I.M)(piano__0=s0.h, violin__1=r + r) + (V % V.M)(piano__0=s1 + s2, violin__1=s3 + s1)
    expected_score = (I % I.M)(piano__0=s0.h) + (V % V.M)(piano__0=s1 + s2, violin__1=s3 + s1)
    assert score.remove_silenced_instruments() == expected_score


def test_project_pattern_voicing():
    from musiclang import Score

    pattern = (I % I.M)(x0 + su1 + su2 + su1, x1.e + bu1.e + bd1 + bu1 + bd1)
    voicing = [b0, b1]
    expected_result = (I % I.M)(piano__0=b0 + su1 + su2 + su1, piano__1=b1.e + bu1.e + bd1 + bu1 + bd1)
    # By voicing
    result_voicing = pattern.project_pattern(voicing)
    assert expected_result == result_voicing

def test_project_pattern_score_restart():
    # By Score
    pattern = (I % I.M)(x0 + su1 + su2 + su1, x1.e + bu1.e + bd1 + bu1 + bd1)
    expected_result = ((I % II.M)(
	piano__0=b0 + su1,
	piano__1=b1.e + bu1.e + bd1)+
    (II % II.M)(
	piano__0=b0 + su1,
	piano__1=b1.e + bu1.e + bd1))

    score_voicing = (I % II.M)(b0.h, b1.h) + (II % II.M)(b0.h, b1.h)

    result_score = pattern.project_pattern(score_voicing, restart_each_chord=True)
    assert expected_result == result_score


def test_project_pattern_score_no_restart():
    # By Score
    pattern = (I % I.M)(x0 + su1 + su2 + su1, x1.e + bu1.e + bd1 + bu1 + bd1)
    expected_result = ((I % II.M)(
	piano__0=b0 + su1,
	piano__1=b1.e + bu1.e + bd1)+
(II % II.M)(
	piano__0=su2 + su1,
	piano__1=bu1 + bd1))

    score_voicing = (I % II.M)(b0.h, b1.h) + (II % II.M)(b0.h, b1.h)

    result_score = pattern.project_pattern(score_voicing, restart_each_chord=False)
    assert expected_result == result_score


def test_project_pattern_score_repeat_if_necessary():
    # By Score
    pattern = (I % I.M)(x0 + su1, x1.e + bu1.e + bd1)
    expected_result = ((I % II.M)(
	piano__0=b0 + su1 + b0 + su1,
	piano__1=b1.e + bu1.e + bd1 + b1.e + bu1.e + bd1))

    score_voicing = (I % II.M)(b0.w, b1.w)

    result_score = pattern.project_pattern(score_voicing, restart_each_chord=False)
    assert expected_result == result_score

from musiclang.library import *

def test_split_too_long_chords():
    score = None + (I %V.M)(s0.augment(7) + s1.augment(2), s4.augment(9))

    splitted_score = score.split_too_long_chords(8)

    print(splitted_score)

    expected_score = (
    (I % V.M)(
        piano__0=s0.augment(frac(7, 1)) + s1,
        piano__1=s4.augment(frac(8, 1)))+
    (I % V.M)(
        piano__0=l,
        piano__1=l)
    )
    assert splitted_score == expected_score


"""Definition of the Roman numeral analysis deep neural network(s)."""

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers


def AugmentedNet(inputs, outputs, blocks=6):
    """Definition of the AugmentedNet architecture.

    Parameters
    ----------
    inputs :
        
    outputs :
        
    blocks :
         (Default value = 6)

    Returns
    -------

    """
    x = []  # (raw) inputs of the network
    xprime = []  # inputs after initial convolutional blocks
    for i in inputs:
        sequenceLength = i.array.shape[1]
        inputFeatures = i.array.shape[2]
        name = i.name.replace("training_", "")
        xi = layers.Input(shape=(sequenceLength, inputFeatures), name=name)
        x.append(xi)
        for i in range(blocks):
            filters = 2 ** (blocks - 1 - i)
            kernel = 2 ** i
            h = layers.Conv1D(filters, kernel, padding="same")(xi)
            h = layers.BatchNormalization()(h)
            h = layers.Activation("relu")(h)
            xi = layers.Concatenate()([xi, h])
        xprime.append(xi)
    if len(x) > 1:
        inputs = layers.Concatenate()(xprime)
    else:
        inputs = xprime[0]
    h = layers.Dense(64)(inputs)
    h = layers.BatchNormalization()(h)
    h = layers.Activation("relu")(h)
    h = layers.Dense(32)(h)
    h = layers.BatchNormalization()(h)
    h = layers.Activation("relu")(h)
    h = layers.Bidirectional(layers.GRU(30, return_sequences=True))(h)
    h = layers.BatchNormalization()(h)
    h = layers.Bidirectional(layers.GRU(30, return_sequences=True))(h)
    h = layers.BatchNormalization()(h)
    y = []
    for output in outputs:
        outputFeatures = output.outputFeatures
        out = layers.Dense(outputFeatures, name=output.shortname)(h)
        y.append(out)
    model = keras.Model(inputs=x, outputs=y)
    return model


def Micchi2020(inputs, outputs):
    """The model by Micchi et al. (2020).

    Parameters
    ----------
    inputs :
        
    outputs :
        

    Returns
    -------

    """

    def DenseNetLayer(x, b, f, n=1):
        """

        Parameters
        ----------
        x :
            
        b :
            
        f :
            
        n :
             (Default value = 1)

        Returns
        -------

        """
        with tf.name_scope(f"denseNet_{n}"):
            for _ in range(b):
                y = layers.BatchNormalization()(x)
                y = layers.Conv1D(
                    filters=4 * f,
                    kernel_size=1,
                    padding="same",
                    data_format="channels_last",
                )(y)
                y = layers.Activation("relu")(y)
                y = layers.BatchNormalization()(y)
                y = layers.Conv1D(
                    filters=f,
                    kernel_size=8,
                    padding="same",
                    data_format="channels_last",
                )(y)
                y = layers.Activation("relu")(y)
                x = layers.Concatenate()([x, y])
        return x

    def PoolingLayer(x, k, s, n=1):
        """

        Parameters
        ----------
        x :
            
        k :
            
        s :
            
        n :
             (Default value = 1)

        Returns
        -------

        """
        with tf.name_scope(f"poolingLayer_{n}"):
            y = layers.BatchNormalization()(x)
            y = layers.Conv1D(
                filters=k,
                kernel_size=1,
                padding="same",
                data_format="channels_last",
            )(y)
            y = layers.Activation("relu")(y)
            y = layers.BatchNormalization()(y)
            y = layers.MaxPooling1D(
                s, s, padding="same", data_format="channels_last"
            )(y)
        return y

    def MultiTaskLayer(h, outputs):
        """

        Parameters
        ----------
        h :
            
        outputs :
            

        Returns
        -------

        """
        y = []
        for output in outputs:
            outputFeatures = output.outputFeatures
            out = layers.Dense(outputFeatures, name=output.shortname)(h)
            y.append(out)
        return y

    _, sequenceLength, inputFeatures = inputs[0].array.shape
    notes = layers.Input(shape=(sequenceLength, inputFeatures))

    x = DenseNetLayer(notes, b=4, f=8, n=1)
    x = PoolingLayer(x, 32, 2, n=1)
    x = DenseNetLayer(x, 4, 5, n=2)
    x = PoolingLayer(x, 48, 2, n=1)

    x = layers.Bidirectional(
        layers.GRU(64, return_sequences=True, dropout=0.3)
    )(x)

    # I don't think we need the TimeDistributed
    # however, leaving as is
    # https://stackoverflow.com/questions/47305618
    # https://github.com/keras-team/keras/issues/11547
    x = layers.TimeDistributed(layers.Dense(64, activation="tanh"))(x)
    y = MultiTaskLayer(x, outputs)

    model = keras.Model(inputs=[notes], outputs=y)
    return model


available_models = {
    "AugmentedNet": AugmentedNet,
    "Micchi2020": Micchi2020,
}



"""A Roman numeral analysis network with synthetic data and additional tasks."""

from os.path import dirname, basename, isfile, join
import glob

__author__ = "Néstor Nápoles López"
__version__ = "1.9.0"

from .inference import get_model, m21Parse, infer_chords




from musiclang import Note
from musiclang import Silence

class Patternator:

    def __init__(self, restart_each_chord=False):
        self.restart_each_chord = restart_each_chord

    def __call__(self, pattern, score):
        if not self.restart_each_chord:
            projected = pattern.project_on_score(score,
                                                 keep_score=True,
                                                 allow_override=True,
                                                 repeat_to_duration=True)

        else:
            projected = None
            for chord in score.chords:
                projected_chord = pattern.project_on_score(chord, keep_score=True,
                                                           allow_override=True,
                                                           repeat_to_duration=True)
                projected += projected_chord
        for c1, c2 in zip(projected.chords, score.chords):
            parts = pattern.parts
            parts = [p for p in parts if p in c2.score.keys()]
            voicing = []
            for idx, part in enumerate(parts):
                voicing.append(c2.score[part].notes[0])

            voicing = list(sorted(voicing, key=lambda note: c1.to_pitch(note)))
            for idx, part in enumerate(parts):
                for id_note, note in enumerate(c1.score[part].notes):
                    if note.type == 'x':
                        current_note = c1.score[part].notes[id_note]
                        replaced = current_note.replace(note, voicing[note.val], octave=False)
                        c1.score[part].notes[id_note] = replaced

        return projected



class TransformGraph:
    """
    Store a directed graph of musical transformations.
    Can be called to returns outputs, Usually returns a single node which is a score

    Basically an implementation of a directed graph, nothing specific to musiclang

    Parameters :
    -----------

    - inputs nodes : (MusicLangWrite | Transform | Filter | Pipeline)*
    - transform nodes : Transform | Filter | Pipeline
    - Nodes can be : ``Score``, ``Melodies``, ``ConcatPipeline``, ``TransformPipeline``, ``ScoreMerger`` or any function with appropriate inputs/outputs

    Examples
    --------
    >>> from musiclang.library import *
    >>> from musiclang.transform.score_merger import ConcatScores
    >>> score1 =  (I % I.M)(s0, s2, s4)
    >>> score2 =  (II % I.M)(s0, s2, s4)
    >>> score3 =  (III % I.M)(s0, s2, s4)
    >>> g = TransformGraph(inputs=['A', 'B', 'C'], outputs='D')
    >>> g.add(output="D", inputs=['A', 'B', 'C'], transfo=ConcatScores())
    >>> g.add('D', ('A', 'B', 'C'), ConcatScores())
    >>> g['D'] = (('A', 'B', 'C'), ConcatScores())
    >>> g(score1, score2, score3)
    (I % I.M)(s0, s2, s4) + (II % I.M)(s0, s2, s4) + (III % I.M)(s0, s2, s4)

    You can add multioutputs transforms :

    >>> g = TransformGraph(inputs=['A', 'B', 'C'], outputs=('D', 'E'))
    >>> g[('D', 'E')] = (('A', 'B'), (lambda a, b : a+b, b)

    """

    def __init__(self, inputs, outputs, nodes=None):
        self.inputs = inputs
        self.names = set()
        self.compute_order = []
        self.names = self.names.union(inputs)
        self.outputs = tuple(outputs) if isinstance(outputs, (list, tuple)) else outputs
        self.nodes = {}
        if nodes is not None:
            self.add_many(nodes)


    @property
    def _multioutput(self):
        return isinstance(self.outputs, (list, tuple))

    def copy(self):
        new_graph = TransformGraph(self.inputs, self.outputs, nodes=dict(self.nodes))
        return new_graph

    def add_many(self, args):
        for arg in args:
            self.add(*arg)

    def __setitem__(self, outputs, value):
        self.add(outputs, value[0], value[1])

    def add(self, outputs, inputs, transfo):
        """
        Add a node in the graph
        Parameters
        ----------
        outputs: tuple or str
        inputs: tuple or list
        transfo : function
                  Function that will take the inputs and returns the outputs

        """
        outputs = tuple(outputs) if isinstance(outputs, (list, tuple)) else outputs
        self.nodes[outputs] = (inputs, transfo)
        if not self.names.issuperset(inputs):
            raise Exception(f'Unexisting inputs {set(inputs) - self.names}.'
                            f' You must define all inputs before defining the node (Only direct acyclic graphs)')
        if isinstance(outputs, tuple):
            for output in outputs:
                self.names.add(output)
        else:
            self.names.add(outputs)
        self.compute_order.append(outputs)

    def _get_compute_graph(self, *args, **parameters):
        if not self.names.issuperset(self.outputs):
            missing_outputs = set(self.outputs) - self.names
            raise Exception(f'Some outputs : "{missing_outputs}" are not defined ')
        compute_graph = {self.inputs[idx]: arg for idx, arg in enumerate(args)}
        kwargs_compute_graph = {key: val for key, val in parameters.items() if key in self.inputs}
        compute_graph = {**kwargs_compute_graph, **compute_graph}
        for outputs in self.compute_order:
            input_names, transfo = self.nodes[outputs]
            inputs = [compute_graph[i] for i in input_names]
            result = transfo(*inputs)
            if isinstance(outputs, (list, tuple)):
                for r, output in zip(result, outputs):
                    compute_graph[output] = r
            else:
                compute_graph[outputs] = result

        return compute_graph

    def __call__(self, *args, **parameters):
        compute_graph = self._get_compute_graph(*args, **parameters)

        if self._multioutput:
            res = tuple([compute_graph[output] for output in self.outputs])
            return res
        else:
            return compute_graph[self.outputs]




def bjorklund_algorithm(steps, pulses):

    steps = int(steps)
    pulses = int(pulses)
    if pulses > steps:
        raise ValueError('pulses must be less or equal to steps')

    pattern = []
    counts = []
    remainders = []
    divisor = steps - pulses
    remainders.append(pulses)
    level = 0
    while True:
        counts.append(divisor // remainders[level])
        remainders.append(divisor % remainders[level])
        divisor = remainders[level]
        level = level + 1
        if remainders[level] <= 1:
            break
    counts.append(divisor)

    def build(level):
        if level == -1:
            pattern.append(0)
        elif level == -2:
            pattern.append(1)
        else:
            for i in range(0, counts[level]):
                build(level - 1)
            if remainders[level] != 0:
                build(level - 2)

    build(level)
    i = pattern.index(1)
    pattern = pattern[i:] + pattern[0:i]
    return pattern


from .composer.auto_composer import AutoComposer
from .arranger import Arranger, ArrangerTrainer
from .composer.harmony import compose_melody_and_harmony, generate_harmony, harmony_to_melody, melody_to_harmony, auto_enrich_melody

__all__ = ['AutoComposer', 'Arranger', 'ArrangerTrainer',
           'compose_melody_and_harmony', 'generate_harmony',
           'harmony_to_melody', 'melody_to_harmony', 'auto_enrich_melody'
           ]


from musiclang import Note, Silence, Continuation



def test_notes():
    note = Note("s", 0, 0, 1)
    assert note.notes == (None + note).notes


def test_scale_pitch():
    note = Note("s", 1, 1, 1)

    assert note.scale_pitch == 8

def test_scale_pitch_negative():
    note = Note("s", -1, 1, 1)

    assert note.scale_pitch == 6

def test_delta_value():
    note = Note("su", 1, 0, 1)
    # FIXME : Should delta value takes in account octaves ?
    assert note.delta_value == 1


def test_is_note():
    assert not Silence(1).is_note
    assert not Continuation(1).is_note
    assert Note("s", 1, 0, 1).is_note

def test_is_silence():
    note = Silence(1)
    assert note.is_silence
    assert not Note("s", 1, 0, 1).is_silence


def test_is_continuation():
    note = Continuation(1)
    assert note.is_continuation
    assert not Note("s", 1, 0, 1).is_continuation


import partitura as pt
import pandas as pd
import warnings
from fractions import Fraction as frac

def _get_bars_from_parts(note_info):
    """
    Get bars start and end beat for each bar of the song
    """
    bars = None
    max_time = -1

    # First get longest part
    longest_part = max(note_info.parts, key=lambda x: len(x.measures))

    barss = []
    for part in note_info.parts:
        starts = []
        ends = []
        for measure in part.measures:
            start, end = measure.start.t, measure.end.t
            starts.append(start)
            ends.append(end)

        beats_start = part.beat_map(starts)
        beats_end = part.beat_map(ends)
        barss.append([(s, e) for s, e in zip(beats_start, beats_end)])

    # Concatenate all bars in barss
    bars = list(sorted(list(set([item for sublist in barss for item in sublist])), key=lambda x: x[0]))
    return bars

import numpy as np
def load_score(midi_file, merge_tracks=True, ticks_per_beat=480):
    import partitura.musicanalysis as analysis
    note_info = pt.load_score_midi(midi_file,
                                   part_voice_assign_mode=5,
                                   quantization_unit=None,
                                   estimate_voice_info=False,
                                   estimate_key=False,
                                   assign_note_ids=True,
                                   )


    bars = _get_bars_from_parts(note_info)
    note_info_2 = pt.load_performance_midi(midi_file)
    array = note_info.note_array()
    dtypes = array.dtype
    columns = list(dtypes.names)
    array_2 = note_info_2.note_array()
    dtypes2 = array_2.dtype
    columns_2 = list(dtypes2.names)



    # Sort by
    assert len(array) == len(array_2), "Partitura does not output the same arrays when load_score_midi and load_performance_midi"

    try:
        assert all(array['pitch'] == array_2['pitch']), "Partitura does not output the same pitches when load_score_midi and load_performance_midi"
    except:
        array = np.sort(array, order=['pitch', 'onset_beat', 'duration_beat'])
        array_2 = np.sort(array_2, order=['pitch', 'onset_sec', 'duration_sec'])
        assert all(array['pitch'] == array_2['pitch']), "After sorting, Partitura does not output the same pitches when load_score_midi and load_performance_midi"

    df = pd.DataFrame(array[columns], columns=columns)
    df2 = pd.DataFrame(array_2, columns=columns_2)
    df['offset_beat'] = df['onset_beat'] + df['duration_beat']
    df['velocity'] = df2['velocity']
    df['track'] = df2['track']
    df['channel'] = df2['channel']
    df = df.sort_values(by=['onset_beat', 'pitch', 'duration_beat'])
    def estimate_voices_with_index(x):
        result = pd.Series(analysis.estimate_voices(x.to_records()))
        result.index = x.index  # Set the index to match the input
        return result
    if merge_tracks:
        df['old_track'] = df['track']
        df['track'] = 0
        gpb = df.groupby('channel').apply(estimate_voices_with_index).reset_index(level=0, drop=True)
        try:
            df['voice'] = gpb
        except:
            df['voice'] = np.squeeze(gpb.values.T)
        df['track'] = df['old_track']
    else:
        df['voice'] = estimate_voices_with_index(df)

    # Quantize onset beats and offset beats
    #df['onset_beat'] = df['onset_beat'].apply(lambda x: float(frac(x).limit_denominator(8)))
    #df['offset_beat'] = df['offset_beat'].apply(lambda x: float(frac(x).limit_denominator(8)))

    notes = df[['onset_beat', 'offset_beat', 'velocity', 'pitch', 'track', 'channel', 'voice']].values.tolist()
    return notes, bars


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

from .voice_separation import separate_voices
from musiclang.write.note import Silence, Continuation
from musiclang.write.constants import OCTAVES
from musiclang import Score

def infer_score_with_chords_durations(sequence, chords, instruments):
    """Get the score from a note sequence, the chords with durations and the instruments assigned to each tracks

    Parameters
    ----------
    sequence :
        param chords:
    instruments :
        return:
    chords :
        

    Returns
    -------

    """
    # Split each chords, instruments, voices
    time_start = 0
    time_end = 0
    score = []
    continuations = {}
    offsets_voices = {}  # Store the offset of voice to deduplicate voice idx between tracks
    offsets_voices_raw = {}  # Store the offset of voice to deduplicate voice idx between tracks
    tracks = list(set([s.track for s in sequence]))
    for channel, instrument in instruments.items():
        for track_idx in tracks:
            track_notes = [n for n in sequence if n.track == track_idx]
            voices = {n.voice for n in track_notes}
            offsets_voices[track_idx] = offsets_voices_raw.get(channel, 0)
            if channel not in offsets_voices_raw.keys():
                offsets_voices_raw[channel] = max(voices) + 1
            else:
                offsets_voices_raw[channel] += max(voices) + 1
    for idx, chord in enumerate(chords):
        time_start = 0 if idx == 0 else time_start + chords[idx - 1].duration
        time_end += chord.duration
        chord_notes = [n for n in sequence if time_start <= n.start < time_end]
        chord_dict = {}
        for track in tracks:
            track_notes = [n for n in chord_notes if n.track == track]
            voices = {n.voice for n in track_notes}
            for voice in voices:
                voice_notes = [n for n in track_notes if n.voice == voice]
                if len(voice_notes) > 0:
                    instrument = instruments.get(voice_notes[0].channel, 'piano')
                    voice_name = instrument + '__' + str(offsets_voices.get(track, 0) + int(voice))
                    cont = continuations.get(voice_name, None)
                    chord_dict[voice_name], cont = _parse_voice(voice_notes, chord,time_start, time_end, 1, cont, is_drum=instrument.startswith('drum'))
                    if cont is not None:
                        continuations[voice_name] = cont

        score.append(chord(**chord_dict))

    return Score(score)


def infer_score(sequence, chords, instruments, bar_duration_in_ticks, offset_in_ticks, tick_value):
    """

    Parameters
    ----------
    sequence :
        
    chords :
        
    instruments :
        
    bar_duration_in_ticks :
        
    offset_in_ticks :
        
    tick_value :
        

    Returns
    -------

    """

    # Split each chord, instrument, voice
    time_start = 0
    time_end = bar_duration_in_ticks
    score = None
    continuations = {}

    # Get all track, voices
    offsets_voices = {}
    offsets_voices_raw = {}
    tracks = list(set([s.track for s in sequence]))
    for channel, instrument in instruments.items():
        for track_idx in tracks:
            track_notes = [n for n in sequence if n.track == track_idx]
            voices = {n.voice for n in track_notes}
            offsets_voices[track_idx] = offsets_voices_raw.get(channel, 0)
            if channel not in offsets_voices_raw.keys():
                offsets_voices_raw[channel] = max(voices) + 1
            else:
                offsets_voices_raw[channel] += max(voices) + 1


    for chord in chords:
        chord_notes = [n for n in sequence if time_start <= n.start < time_end]
        chord_dict = {}
        for track in tracks:
            track_notes = [n for n in chord_notes if n.track == track]
            voices = {n.voice for n in track_notes}
            for voice in voices:
                voice_notes = [n for n in track_notes if n.voice == voice]
                if len(voice_notes) > 0:
                    instrument = instruments[voice_notes[0].channel]
                    voice_name = instrument + '__' + str(offsets_voices[track] + int(voice))
                    cont = continuations.get(voice_name, None)
                    chord_dict[voice_name], cont = _parse_voice(voice_notes, chord,
                                                            time_start, time_end, tick_value, cont)
                    if cont is not None:
                        continuations[voice_name] = cont
                    chord_dict[voice_name] = chord_dict[voice_name].o(- OCTAVES.get(instrument, 0))


        score += chord(**chord_dict)
        time_start += bar_duration_in_ticks
        time_end += bar_duration_in_ticks

    return score



def infer_voices_per_tracks(sequence):
    """Separate voices of each tracks

    Parameters
    ----------
    sequence :
        return: sequence_result: New sequence with assigned "voice" property

    Returns
    -------
    type
        sequence_result: New sequence with assigned "voice" property

    """
    # Get all tracks
    tracks = list(set([int(s.track) for s in sequence]))
    # Separate voices for each tracks
    sequence_result = []
    for track in tracks:
        # Get notes
        notes = [s.array() for s in sequence if s.track == track]
        new_notes = separate_voices(notes)
        sequence_result += new_notes

    sequence_result = list(sorted(sequence_result, key=lambda x: x.start))
    return sequence_result


"""
Utils
"""


def _parse_voice(voice_notes, chord, bar_time_start, bar_time_end, tick_value, cont, is_drum=False):
    """Parse a single voice to a musicLang melody between start time and end time

    Parameters
    ----------
    voice_notes :
        list of notes from sequence that represents a voice
    chord :
        chord on which to parse the melody (to get the references pitches)
    bar_time_start :
        Time start
    bar_time_end :
        Time end
    tick_value :
        Unit of time
    cont :
        Is there a continuation to take in account and to add to melody

    Returns
    -------
    type
        melody : musiclang.Melody : parsed melody,

    """
    def _parse_note(note, duration, chord, tick_value):
        """

        Parameters
        ----------
        note :
            
        duration :
            
        chord :
            
        tick_value :
            

        Returns
        -------

        """
        value = chord.parse(note.pitch - 60)
        value = value.augment(duration)
        value.amp = note.vel
        return value.augment(tick_value)

    melody = []
    return_cont = None
    local_time_end = voice_notes[0].start
    if cont is not None:
        if cont.duration > 0:
            melody.append(cont)
        local_time_end = bar_time_start + (cont.duration / tick_value)

    elif local_time_end > bar_time_start:
        duration = (local_time_end - bar_time_start) * tick_value
        if duration > 0:
            melody.append(Silence(duration))

    for idx, note in enumerate(voice_notes):
        if is_drum:
            if idx < len(voice_notes) - 1:
                next_note = voice_notes[idx + 1]
                note.end = next_note.start
            else:
                note.end = bar_time_end

        overlap = local_time_end - note.start
        if overlap > 0:
            melody[-1].duration -= overlap * tick_value
            if melody[-1].duration == 0:
                melody.pop()
            duration = note.end - note.start
            n = _parse_note(note, duration, chord, tick_value)
            if n.duration > 0:
                melody.append(n)
        elif overlap < 0:
            melody.append(Silence(- overlap * tick_value))
            duration = note.end - note.start
            if duration > 0:
                n = _parse_note(note, duration, chord, tick_value)
                if n.duration > 0:
                    melody.append(n)
        else:
            duration = note.end - note.start
            if duration > 0:
                n = _parse_note(note, duration, chord, tick_value)
                if n.duration > 0:
                    melody.append(n)

        local_time_end = note.end
        # Find scale note
    if local_time_end < bar_time_end:
        melody.append(Silence((bar_time_end - local_time_end) * tick_value))
    if local_time_end > bar_time_end:
        return_cont = Continuation((local_time_end - bar_time_end) * tick_value)
        melody[-1].duration -= (local_time_end - bar_time_end) * tick_value
        if melody[-1].duration == 0:
            melody.pop()

    from musiclang import Note, Melody
    melody = Melody(melody)
    assert all([isinstance(m, Note) for m in melody.notes])
    delta_t = abs(melody.duration -  ((bar_time_end - bar_time_start) * tick_value))
    assert delta_t < 0.25, f"Issue with melody duration {delta_t}"
    assert melody.notes[0].duration >0, "Issue with note duration"
    return melody, return_cont







"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

class NotePitch:
    """ """

    def __init__(self, pitch, offset=0, duration=1, velocity=120, track=0, silence=0, continuation=0,
                 tempo=None, pedal=None
                 ):
        self.pitch = pitch
        self.offset = offset
        self.duration = duration
        self.velocity = velocity
        self.track = track
        self.silence = silence
        self.continuation = continuation
        self.tempo = tempo
        self.pedal = pedal


    def is_note(self):
        """ """
        return not (self.silence or self.continuation)

    def to_midi_note(self):
        """ """
        return [self.pitch, self.offset, self.duration, self.velocity, self.track, self.silence, self.continuation,
                self.tempo, self.pedal
                ]






from musiclang.library import *
from .score_formatter_elements import BarLine, BarChord, Beat, MultiBar, Event, TimeSignature, TonalityLine
import re
"""
The roman numeral parsed language is an extension of this paper : 
https://archives.ismir.net/ismir2019/paper/000012.pdf

By Dmitri Tymoczko Mark Gotham Michael Scott Cuthbert and Christopher Ariza

"""
DEFAULT_INSTRUMENTS = ['piano__0', 'piano__1', 'piano__2', 'piano__3']
DEFAULT_VOICING = (b0, b1, b2, b3)
class ScoreFormatter:

    def __init__(self, text, instruments=None, voicing=None):
        self.text = text
        self.time_signature = (4, 4)
        self.prev_time_signature = (4, 4)
        self.elements = []
        self.first_change_time_signature = True
        self.bar_elements = {}
        self.allow_multi_signature = True
        ## PARAMETERS
        self.rhythm = {}
        self.counterpoint = []
        self.voice_leading = []
        self.chord_idx_voice_leading = 0
        self.chord_idx_counterpoint = 0
        self.key = 0
        self.init_offset_bar = 0
        self.mode = 'M'
        self.instruments = DEFAULT_INSTRUMENTS if instruments is None else instruments
        self.init_bar_number = -1
        self.current_score = DEFAULT_VOICING if voicing is None else voicing
        self.bar_number = 0
        self.chord_number = 0
        self.current_beat = 0
        self.pickup = 0
        self.chord_started_time = (0, 0)  # Bar idx, beat idx
        ## Init elements
        self.init()

    @classmethod
    def convert_harmony_with_relative_tones(cls, text):
        """
        Convert roman numeral harmony to roman numeral with only one tonality at the beginning
        Parameters
        ----------
        text

        Returns
        -------

        """



    def add_chord(self, chord, score):
        if score is not None:
            # Deduce how much last chord should lasts
            bar_start, beat_start = self.chord_started_time
            duration = self.prev_duration * (self.bar_number - bar_start) + (self.current_beat - beat_start)
            if duration == 0:
                score.chords = score.chords[:-1]
            else:
                score.chords[-1] = score.chords[-1].set_duration(duration)
                assert score.chords[-1].duration == duration
        # Create new chord started time
        self.chord_started_time = (self.bar_number, self.current_beat)
        if score is None and self.current_beat > 0:
            self.pickup = self.current_beat
        score += chord
        return score

    def set_voice_leading(self, voice_leading):
        if len(self.voice_leading) > 0:
            self.voice_leading[-1][1] = self.chord_number
        self.voice_leading = self.voice_leading + [[self.chord_number, None, voice_leading]]

    def set_counterpoint(self, fixed_voices):
        if len(self.counterpoint) > 0:
            self.counterpoint[-1][1] = self.chord_number
        self.counterpoint = self.counterpoint + [[self.chord_number, None, fixed_voices]]

    def set_rhythm(self, instrument, nb_bars, rhythm):
        from musiclang import Metric
        metric = (rhythm, nb_bars)
        if instrument in self.rhythm.keys():
            self.rhythm[instrument][-1][1] = self.chord_number
        self.rhythm[instrument] = self.rhythm.get(instrument, []) + [[self.chord_number, None, metric]]

    def set_time_signature(self, time_signature):
        if self.time_signature is not None:
            if not self.allow_multi_signature:
                raise Exception('Cannot handle multi signature scores with "allow_multi_signature" flag to false')
            self.prev_time_signature = tuple(self.time_signature)
        else:
            self.prev_time_signature = time_signature

        if self.first_change_time_signature:
            self.first_change_time_signature = False
            self.prev_time_signature = time_signature

        self.time_signature = time_signature

    def set_bar_number(self, idx):
        assert (idx > self.bar_number) or (self.bar_number == 0)

        self.bar_number = idx
        self.current_beat = 0

    def set_current_beat(self, beat):
        if beat <= self.current_beat:
            pass
        else:
            self.current_beat = beat

    def __iter__(self):
        for element in self.elements:
            yield element

    @property
    def duration(self):
        if self.time_signature is None:
            return 4
        return 4 * self.time_signature[0] / self.time_signature[1]

    @property
    def prev_duration(self):
        if self.prev_time_signature is None:
            return frac(4 * self.time_signature[0] / self.time_signature[1]).limit_denominator(8)
        return frac(4 * self.prev_time_signature[0] / self.prev_time_signature[1]).limit_denominator(8)


    @property
    def base(self):
        return 4 / self.time_signature[1]

    def is_time_signature(self, line):
        return line.startswith('Time') or line.startswith('time')

    def is_tonality(self, line):
        return line.lower().startswith('tonality') or line.lower().startswith('key')
    def is_event(self, line):
        return line.startswith('!')

    def is_bar(self, line):
        return line.startswith('m')

    def is_multibar(self, line):
        first_word = line.split(' ')[0]
        return line.startswith('m') and (('-' in first_word) or ('=' in line))

    def apply_voice_leading(self, score):
        # Normalize score
        for start, end, voice_leading in self.voice_leading:
            score[start: end] = voice_leading(score[start: end].normalize_instruments())
        return score

    def apply_counterpoint(self, score):
        from musiclang import Melody
        # Normalize score
        score = score.normalize_instruments()
        for start, end, fixed_voices in self.counterpoint:
            # Find first notes of each ...
            # Apply counterpoint
            from musiclang.transform.library import create_counterpoint_on_score
            counterpoint_score = score[start:end].copy()
            if len(fixed_voices) == 0:
                # Add a fake bass
                for chord in counterpoint_score.chords:
                    chord.score['temp_bass__0'] = b0.o(-2).augment(chord.duration)
                fixed_voices = ['temp_bass__0']
            counterpoint_score = counterpoint_score.to_standard_note()
            counterpoint_score = create_counterpoint_on_score(counterpoint_score, fixed_parts=fixed_voices)
            # Remove fake bass
            # Reapply first notes of each chord to keep harmonic content ok
            for idx, chord in enumerate(counterpoint_score):
                for ins, melody in chord.score.items():
                    if ins in score[start + idx].score.keys():
                        score[start + idx].score[ins] = score[start + idx].score[ins].notes[0] + Melody([n for n in melody.notes[1:]])

        return score

    def apply_rhythm(self, score):
        from musiclang import ScoreRhythm
        from musiclang import Metric
        for key, rhythm_array in self.rhythm.items():
            for start_idx, end_idx, metric_data in rhythm_array:
                rhythm, nb_bars = metric_data
                metric = Metric.FromArray(rhythm, signature=self.time_signature, nb_bars=nb_bars)
                score[start_idx:end_idx] = ScoreRhythm(rhythm_dict={key: metric})(score[start_idx: end_idx])
        return score

    def parse(self, allow_multi_signature=True, tonality=None):
        """
        Parse the chord progression into music lang using voice leading
        """

        if tonality is not None:
            for idx, el in enumerate(self.elements):
                if isinstance(el, TonalityLine):
                    self.elements[idx] = TonalityLine("Tonality :" + tonality)

        self.allow_multi_signature = allow_multi_signature
        interpreter = ScoreInterpreter(self)
        score = interpreter.parse()
        if self.voice_leading is not None:
            score = self.apply_voice_leading(score)
        if self.rhythm:
            # Apply proper rhythm to each section
            score = self.apply_rhythm(score)
        if self.counterpoint is not None:
            # Apply counterpoint to notes starting from second if voice leading
            score = self.apply_counterpoint(score)
        score.config['pickup'] = self.pickup
        score.config['time_signature'] = self.time_signature
        score.config['annotation'] = self.text
        return score

    @classmethod
    def init_object(cls, elements):
        return cls(''.join([e.to_text() for e in elements]))

    @classmethod
    def get_score(cls, elements):
        return cls.init_object(elements).parse()

    def init(self):
        lines = self.text.split('\n')
        for line in lines:
            line = line.lstrip('\t').lstrip(' ')
            if self.is_time_signature(line):
                self.elements.append(TimeSignature(line))
            elif self.is_tonality(line):
                self.elements.append(TonalityLine(line))
            elif self.is_event(line):
                self.elements.append(Event(line, self))
            elif self.is_multibar(line):
                multibar = MultiBar(line)
                new_range = multibar.receiver_range
                old_range = multibar.emitter_range
                # Copy proper elements
                for idx, old_idx in zip(range(*new_range), range(*old_range)):
                    self.elements.append(BarLine(f"m{idx}", self))
                    self.elements += self.bar_elements[old_idx]
                    self.bar_elements[idx] = self.bar_elements[old_idx]
                    self.chord_number += len([e for e in self.bar_elements[idx] if isinstance(e, BarChord)])
            elif self.is_bar(line):
                bar = BarLine(line, self)
                elements = bar.get_elements()
                if self.init_bar_number > bar.idx:
                    continue
                    #raise Exception('Invalid bar number, should be higher than current')
                elif self.init_bar_number != bar.idx:  # Get only first variation of a bar
                    self.elements.append(bar)
                    self.elements += elements
                    self.init_bar_number = bar.idx
                    self.bar_elements[bar.idx] = elements
                    self.chord_number += len([e for e in elements if isinstance(e, BarChord)])

    def __repr__(self):
        return f"ScoreFormatter({self.elements})"

class ScoreInterpreter:

    def __init__(self, sf):
        self.sf = sf

    def parse(self):
        score = None
        for event in self.sf:
            score = event.parse(score, self.sf)

        return score



"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from functools import cached_property

from .constants import *
import re

class Chord:
    """
    Represents a chord in MusicLang.

    The chord is composed of :

        - One degree in ``I, II, III, IV, V, VI, VII``
        - One extension (optional) that represents figured bass in ``5,6,64,7,65,43,2,9,11,13``
        - One tonality itself composed of :

            - A degree in ``I, II, III, IV, V, VI, VII``
            - A mode in ``'M', 'm', 'mm', 'dorian, 'phrygian', 'lydian', 'mixolydian', 'aeolian', 'locrian'``
        - One score : A Dictionary where keys are part names, values are :func:`~musiclang.Melody`
        - One octave : An integer that specifies on which octave transpose the chord

    .. note:: In reality a chord in MusicLang is a "chord scale" because it always implies a tonality.

    Examples
    --------

    >>> from musiclang.library import *
    >>> score =  (I % I.M)(s0, s2)
    >>> score
    (I % I.M)(
        piano_0=s0,
        piano_1=s2)

    You can also pass a list of melodies to an instrument :

    >>> score =  (I % I.M)(violin=[s0, s2, s4.e + s2.e])
    >>> score
    (I % I.M)(
        violin__0=s0,
        violin__1=s2,
        violin__2=s4.e + s2.e)

    Finally can also pass a list of voices to an instrument :

    >>> score =  (I % I.M)(piano__0=s0 + s2, piano__2= s2 + s4)
    >>> score
     (I % I.M)(
         piano__0=s0 + s2,
         piano__2= s2 + s4)

    """

    EXCLUDED_ITEMS = ['__array_struct__']

    def __init__(self, element, extension='', tonality=None, score=None, octave=0, tags=None):
        """

        Parameters
        ----------
        element: int
        extension: int or str , optional
                 Represents the figured bass. Must be in ['','5','6','64','7','65','43','2','9','11','13']
        tonality: musiclang.Tonality, optional
                Represents the tonality of the chord
        score: dict, optional
               Represents the melodies and instruments played by this chord
        octave: int, optional
                Represents the base octave of the chord
        """
        self.element = element
        self.extension = str(extension)
        self.extension = self.normalize_extension()
        self.tonality = tonality
        self.octave = octave
        self.score = {} if score is None else score
        self.tags = set(tags) if tags is not None else set()

    def has_tag(self, tag):
        """
        Check if the tag exists for this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        return tag in self.tags

    def add_tag(self, tag):
        """
        Add a tag to this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags.add(tag)
        return cp


    def transpose(self, interval):
        cp = self.copy()
        cp.tonality.degree += interval
        return cp

    def add_tags(self, tags):
        """
        Add several tags to the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        chord: Chord

        """
        cp = self.copy()
        cp.tags = cp.tags.union(set(tags))
        return cp

    def clear_note_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        for k, v in cp.score.items():
            cp.score[k] = v.clear_note_tags()
        return cp

    def remove_tags(self, tags):
        """
        Remove several tags from the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]

        Returns
        -------
        chord: Chord


        """
        cp = self.copy()
        cp.tags = cp.tags - set(tags)
        return cp

    def remove_tag(self, tag):
        """
        Remove a tag from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags.remove(tag)
        return cp

    def clear_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags = set()
        return cp

    def has_pitch(self, pitch):
        """
        Check whether the pitch belong to the chord or not

        Parameters
        ----------
        pitch: int
        Pitch to check

        Returns
        -------
        result: boolean
            True if the chord has this pitch

        """
        return (pitch % 12) in [p % 12 for p in self.chord_pitches]

    def parse(self, pitch):
        """
        Parse an integer pitch (0=C5)
        Parameters
        ----------
        pitch :
            

        Returns
        -------

        """
        from .note import Note
        if pitch % 12 in self.scale_set:
            scale = self.scale_pitches
            type = 's'
        else:
            scale = self.chromatic_scale_pitches
            type = 'h'

        scale_mod = [s % 12 for s in scale]
        idx = scale_mod.index(pitch % 12)
        oct = (pitch - scale[0]) // 12
        return Note(type, idx, oct, 1)

    def project_on_score(self, score2, **kwargs):
        """
        Transform chord into a score and apply Score.project_on_score
        Parameters
        ----------
        score2: other score on which to project
        voice_leading:
        kwargs
        Returns
        -------
        """
        return self.to_score().project_on_score(score2, **kwargs)

    def decompose_duration(self):
        """ """
        return self(**{key: melody.decompose_duration() for key, melody in self.score.items()}, tags=self.tags)

    @property
    def degree(self):
        return self.element

    def get_note_degree_accident(self, note):
        from .constants import DEGREE_TO_SCALE_DEGREE
        if note.type == 's':
            return note.val
        elif note.type in ['l', 'r']:
            return (None, None)
        elif note.type == 'h':
            pass


    def realize_tags(self, last_note=None, final_note=None):
        new_chord_dict = {}
        last_note = last_note or {}
        final_note = final_note or {}
        for key, melody in self.score.items():
            new_chord_dict[key] = melody.realize_tags(last_note.get(key, None), final_note.get(key, None))
        return self(**new_chord_dict)


    def correct_chord_octave(self):
        """
        Correct chord octaves to minimize distance with central C

        Returns
        -------
        score: Score
        """
        from musiclang.analyze import inverse_recursive_correct_octave
        return inverse_recursive_correct_octave(self)

    def split(self, max_length):
        """
        Split the chords that are too long into smaller chords
        Parameters
        ----------
        max_length: fraction
           The maximum length of a chord

        Returns
        -------
        score: Score
           The score with shorter chords

        """
        score = None
        if self.duration <= max_length:
            return self
        nb_chords = 1 + (self.duration // max_length)
        remaining_duration = self.duration % max_length
        from musiclang.library import s0
        chords = [self(s0.augment(max_length)) for i in range(nb_chords)]
        if remaining_duration > 0:
            chords[-1].score['piano__0'] = chords[-1].score['piano__0'].set_duration(remaining_duration)
        else:
            chords = chords[:-1]
        chords = sum(chords, None)
        return self.project_on_score(chords)


    @property
    def mode(self):
        return self.tonality.mode

    def change_mode(self, mode):
        """

        Parameters
        ----------
        mode :
            

        Returns
        -------

        """
        new_chord = self.copy()
        new_chord.tonality = new_chord.tonality.change_mode(mode)
        return new_chord

    @cached_property
    def top_note(self):
        return self.extension_notes[-1]

    @cached_property
    def lowest_note(self):
        return self.extension_notes[0]

    def middle_note(self, index):
        if len(self.extension_notes) == 1:
            return self.extension_notes[0]
        if len(self.extension_notes) == 2:
            return self.extension_notes[0]
        if len(self.extension_notes) == 3:
            return self.extension_notes[1]
        if len(self.extension_notes) == 4:
            return self.extension_notes[index+1]
        else:
            return self.extension_notes[index + 1]

    @cached_property
    def middle_note1(self):
        return self.middle_note(0)

    @cached_property
    def middle_note2(self):
        return self.middle_note(1)


    @property
    def pedal(self):
        """
        Apply pedal on first note and release on last
        """
        return self(**{key: item.to_melody().pedal for key, item in self.score.items()}, tags=set(self.tags))

    def to_pitch(self, note, last_pitch=None):
        """

        Parameters
        ----------
        note :
            
        last_pitch :
             (Default value = None)

        Returns
        -------

        """
        from .out.to_midi import note_to_pitch_result
        if note.is_continuation:
            return last_pitch

        if not note.is_note:
            return None
        return note_to_pitch_result(note, self, last_pitch=last_pitch)

    def to_absolute_note(self):
        return self(**{part: melody.to_absolute_note(self) for part, melody in self.score.items()})

    @cached_property
    def bass_pitch(self):
        return self.chord_extension_pitches[0]

    def to_sequence(self):
        """
        See :func:`~Score.to_sequence()
        """
        sequence_dict = []
        for inst in self.instruments:
            sequence_dict += self.score[inst].to_sequence(self, inst)

        return sequence_dict

    def to_score(self, copy=True):
        """ """
        from .score import Score
        chord = self if not copy else self.copy()
        return Score([chord])

    def show(self, *args, **kwargs):
        """
        See :func:`~Score.show()
        """
        return self.to_score().show(*args, **kwargs)


    def to_musicxml(self, *args, **kwargs):
        return self.to_score().to_musicxml(*args, **kwargs)


    def to_music21(self, *args, **kwargs):
        return self.to_score().to_music21(*args, **kwargs)


    @cached_property
    def scale_degree(self):
        """

        Returns
        -------

        """
        tonic = self.scale_pitches[0]
        return DEGREE_TO_SCALE_DEGREE[tonic % 12] + tonic // 12

    @cached_property
    def pitch_set(self):
        """
        Get a frozenset of :func:`~Chord.chord_pitches`
        """
        return frozenset({s % 12 for s in self.chord_pitches})

    @property
    def pitch_dict(self):
        chromatic_dict = {self.to_pitch(note): note for note in self.chromatic_scale_notes}
        diatonic_dict = {self.to_pitch(note): note for note in self.scale_notes}

        return {**chromatic_dict, **diatonic_dict}


    @cached_property
    def scale_set(self):
        """
        Get a frozenset of :func:`~Chord.scale_pitches`
        """
        return frozenset({s % 12 for s in self.scale_pitches})

    @cached_property
    def chord_pitches(self):
        """
        Get the chord absolute pitches (integers) starting with the chord root.

        See Also
        --------
        :func:`~Tonality.scale_pitches`
        :func:`~Tonality.chord_extension_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> (I % I.M).chord_pitches
        [0, 4, 7]

        >>> from musiclang.library import *
        >>> (I % I.M)['6'].chord_pitches
        [0, 4, 7]
        """
        notes = self.chord_notes
        return [self.to_pitch(n) for n in notes]

    def remove_accidents(self):
        return Chord(**{key: val.remove_accidents() for key, val in self.score.items()}, tags=set(self.tags))

    def predict_score(self, **kwargs):
        """
        Predict the continuation of the chord/score with a LLM model.
        See :func:`~Score.predict_score()`
        """
        return self.to_score().predict_score(**kwargs)

    def apply_pattern(self, chords, instruments=None, voicing=None, restart_each_chord=False,
                      fixed_bass=True, voice_leading=True, amp='mf'):
        """
        Apply a chord progression to a pattern
        Parameters
        ----------
        pattern: Score or Chord
            Pattern to use
        chords: Score or list or Chord
            Chord progression to use
        instruments: list[str] or None
            Instruments to use, by default use only piano
        voicing: list[Note] or None
            Voicing to use in the pattern. By default, spread the voicing between root -2 octave and fifth +1 octave
        restart_each_chord: bool
            If True, the pattern will be restarted at each chord
        fixed_bass: bool
            If True, the bass will be fixed in the voice leading
        voice_leading: bool
            If True, the voice leading optimizer will be applied
        amp: str
            Amplitude of the pattern (in ppp, pp, p, mp, mf, f, ff, fff)

        Returns
        -------
        Score

        """
        from musiclang.predict.composer import apply_pattern
        return apply_pattern(self, chords, instruments=instruments, voicing=voicing, restart_each_chord=restart_each_chord,
                      fixed_bass=fixed_bass, voice_leading=voice_leading, amp=amp)

    def normalize_chord_duration(self):
        """
        Normalize the duration of the chord (project on the melody with the shortest duration)
        See :func:`~Score.normalize_chord_duration()`
        """
        min_duration = min([melody.duration for melody in self.score.values()])
        base_chord = self.to_chord().set_duration(min_duration)
        return self.project_on_score(base_chord).chords[0]

    def get_scale_from_type(self, type):
        if type == "h":
            return self.chromatic_pitches
        elif type == "s":
            return self.scale_pitches
        elif type == "b":
            return self.chord_extension_pitches
        elif type == "c":
            return self.chord_pitches
        else:
            raise ValueError("This type is not associated to a scale")
    @cached_property
    def chromatic_pitches(self):
        return [self.scale_pitches[0] + i for i in range(12)]
    @cached_property
    def scale_pitches(self):
        """
        Get the scale absolute pitches (integers) starting with the chord root.

        See Also
        --------
        :func:`~Tonality.scale_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> (I % I.M).scale_pitches
        [0, 2, 4, 5, 7, 9, 11]

        """
        from .tonality import Tonality
        if self.tonality is None:
            return (self % Tonality(0)).scale_pitches
        tonality_scale_pitches = self.tonality.scale_pitches
        start_idx = self.element
        scale_pitches = tonality_scale_pitches[start_idx:] + [t + 12 for t in tonality_scale_pitches[:start_idx]]
        scale_pitches = [n + 12 * self.octave for n in scale_pitches]
        return scale_pitches


    @cached_property
    def chromatic_scale_pitches(self):
        """ """
        return [self.scale_pitches[0] + i for i in range(12)]

    @cached_property
    def chord_extension_pitches(self):
        """
        Get the chord absolute pitches (integers) starting with the chord bass as indicated by the extension.

        See Also
        --------
        :func:`~Tonality.chord_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> (I % I.M)['6'].chord_extension_pitches
        [4, 7, 12]

        >>> from musiclang.library import *
        >>> (I % I.M)['2[add6]'].chord_extension_pitches
        [-1, 0, 2, 4, 9]
        """

        notes = self.extension_notes
        return [self.to_pitch(n) for n in notes]

    @property
    def chord_extension_pitch_classes(self):
        return [i % 12 for i in self.chord_extension_pitches]
    @property
    def parts(self):
        """
        Get the list of all parts names
        Same as :func:`~Chord.instruments`
        """
        return list(self.score.keys())

    def to_drum(self):
        return self(**{key: val.to_drum() if key.startswith('drum') else val for key, val in self.items()})

    @property
    def instruments(self):
        """
        Get the list of all part names
        """
        return list(self.score.keys())

    def get_part(self, item):
        """
        Get the corresponding part of the chord score. Can be an integer (part index) or a string (part name)

        Parameters
        ----------
        item : str or int
               Part name or part index

        Returns
        -------
        melody: Melody
                The melody corresponding to the part

        """
        if isinstance(item, int):
            return self.score[self.parts[item]]
        else:
            return self.score[item]

    def score_equals(self, other):
        """
        Check if the parts of both chords are equal

        Parameters
        ----------
        other : Chord

        Returns
        -------
        res : bool

        """
        if not isinstance(other, Chord):
            return False
        return self.score == other.score

    def chord_equals(self, other):
        """
        Check if two chords are equal

        Parameters
        ----------
        other :
            

        Returns
        -------
        res: bool

        """
        if not isinstance(other, Chord):
            return False
        return self.element == other.element and self.extension == other.extension and self.tonality == other.tonality and self.octave == other.octave

    def __eq__(self, other):
        from .score import Score
        if isinstance(other, Score):
            return Score([self]) == other
        if not isinstance(other, Chord):
            return False
        return self.chord_equals(other) and self.score_equals(other)

    def __and__(self, other):
        if isinstance(other, int):
            return self(**{part: melody & other for part, melody in self.score.items()})
        else:
            raise Exception(f'Not compatible type with & {other.__class__}')

    def __hash__(self):
        return hash(self.__repr__())

    @property
    def possible_notes(self):
        """
        List the notes that belong to the chord (only the chord, not the scale)

        Returns
        -------
        notes: List[Note]
               The list of possible notes for a chord

        """
        from .note import Note
        if self.extension in ['5', '', '6', '64']:
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1)]
        elif self.extension in ['7', '65', '43', '2']:
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1)]
        elif self.extension == '9':
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1), Note("s", 1, 1, 1)]
        elif self.extension == '11':
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1), Note("s", 1, 1, 1),
                    Note("s", 3, 1, 1)]
        elif self.extension == '13':
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1),
                    Note("s", 1, 1, 1), Note("s", 3, 1, 1), Note("s", 5, 1, 1)]
        else:
            raise Exception('Unknown extension : ' + self.extension)

    @property
    def scale_notes(self):
        """
        List the notes that belong to the chord scale

        Returns
        -------
        notes: List[Note]
               The list of possible notes for the chord scale
        """
        from .note import Note
        return [Note("s", i, 0, 1) for i in range(7)]

    @property
    def chromatic_scale_notes(self):
        """
        List the chromatic notes that belong to the chord scale

        Returns
        -------
        notes: List[Note]
               The list of possible chromatic notes for the chord scale
        """
        from .note import Note
        return [Note("h", i, 0, 1) for i in range(12)]

    @property
    def scale_dissonances(self):
        """
        Returns the notes that don't belong to the chord but belong to the chord scale

        Returns
        -------
        notes: List[Note]
               The list of possible notes for the chord scale
        """
        consonances = {(note.val % 7) for note in self.possible_notes}
        return [note for note in self.scale_notes if note.val not in consonances]

    def __getitem__(self, item):
        """
        Attach a figured bass to the chord (eg : 64)

        Parameters
        ----------

        item: int or str in ['5', '7', '9', '11', '13', '65', '43', '2', '6', '64', '']
              Extension of the chord

        Returns
        -------

        chord: Chord
               A new chord with the figured bass extension

        """
        item = str(item)
        res = self.copy()
        res.extension = str(item)
        try:
            n = res.extension_notes
        except Exception as e:
            raise e
            #raise ValueError(f'Could not parse extension : {item} {e}')
        res.extension = res.normalize_extension()

        return res

    #
    # def __getattr__(self, item):
    #     chord = self.copy()
    #     chord.score = {k: getattr(s, item) for k, s in self.score.items()}
    #     return chord

    @property
    def duration(self):
        """
        Get the chord duration as the max duration of the melodies contained in the chords

        Returns
        -------
        duration: fractions.Fraction or int
                  duration of the chord
        """
        if len(self.score.keys()) == 0:
            return 0
        return max([self.score[key].duration for key in self.score.keys()])

    def set_amp(self, amp):
        return self(**{key: val.set_amp(amp) for key, val in self.items()}, tags=set(self.tags))

    def set_part(self, part, melody, inplace=False):
        """
        Set a part on the chord (inplace operation)
        Parameters
        ----------
        part : part name

        melody :  melody to apply to the part
        inplace: bool
                 If the operation is inplace

        Returns
        -------

        """

        if inplace:
            score = self
        else:
            score = self.copy()
        part_name = score.parts[part] if isinstance(part, int) else part
        score[part_name] = melody
        return score


    def has_score(self):
        """

        """
        return self.score is not None

    def to_chord(self, duration=False):
        """
        Returns a copy of the chord without the chord score

        Returns
        -------
        chord: Chord
               Chord with empty score

        """
        result = self.copy()
        result.score = {}
        if duration:
            result = result.set_duration(self.duration)
        return result

    @property
    def chords(self):
        """
        Returns a Score object with this chord

        Returns
        -------
        score: Score
               The score with this chord as the only element

        """
        from .score import Score
        return Score([self]).chords

    def __mul__(self, other):
        """
        If other is Integer, repeat the note other times
        """
        from .score import Score
        if isinstance(other, int):
            return Score([self.copy() for i in range(other)])
        else:
            raise Exception('Cannot multiply Chord and ' + str(type(other)))

    def __mod__(self, other):
        """
        Modulate to another tonality relative to the current tonality

        Parameters
        ----------
        other: Tonality
            Tonality on which to modulate this chord
        """
        from .tonality import Tonality
        if isinstance(other, Tonality):
            chord = self.copy()
            # Tonality is always referenced from base tonality which is the most outside tonality
            other = other.copy()
            other.octave += self.octave
            chord.tonality = chord.tonality + other
            chord.octave = 0
            return chord
        else:
            raise Exception('Following % should be a Tonality')

    def modulate(self, other):
        """
        Modulate to another tonality relative to the current tonality

        Parameters
        ----------
        other: Tonality
            Tonality on which to modulate this chord
            

        Returns
        -------
        chord: Chord
               Modulated chord

        """
        from .tonality import Tonality
        if isinstance(other, Tonality):
            chord = self.copy()
            # Tonality is always referenced from base tonality which is the most outside tonality
            other = other.copy()
            other.octave += self.octave
            chord.tonality = chord.tonality + other
            chord.octave = 0
            return chord
        else:
            raise Exception('Following % should be a Tonality')

    def __add__(self, other):
        """
        Add a new chord to this chord, it will returns a score

        Parameters
        ----------

        other: Chord or Score
               Create a new score concatenating this and the other object

        Examples
        ---------

        >>> from musiclang.library import *
        >>> chord1 = (I % I.M)(piano__0=s0)
        >>> chord2 = (V % I.M)(piano__0=s0)
        >>> chord + chord2
        (I % I.M)(piano__0=s0) + (V % I.M)(piano__0=s0)

        """

        from .score import Score
        if other is None:
            return self.copy()
        if isinstance(other, Chord):
            return Score([self.copy(), other.copy()])
        if isinstance(other, Score):
            return Score([self.copy()] + other.copy().chords)

    def __radd__(self, other):
        from .score import Score
        if other is None:
            return Score([self.copy()])

    def element_to_str(self):
        """
        Convert the element of the chord into a string

        Returns
        -------
        res: str

        """
        return ELEMENT_TO_STR[self.element]

    def extension_to_str(self):
        """
        Converts the extension of the chord into a string

        Returns
        -------
        res: str

        """
        if str(self.extension) == '5' or str(self.extension) == '':
            return ''
        else:
            return f"['{self.extension}']"


    def tonality_to_str(self):
        """
        Convert the tonality of the chord to a string.

        Parameters
        ----------

        Returns
        -------

        """
        if self.tonality is None:
            return ""

        return " % " + self.tonality.to_code()

    def d(self):
        """Chord up one octave (=o(-1))"""
        return self.o(-1)

    def u(self):
        """Chord down one octave (=o(-1))"""
        return self.o(1)

    def to_extension_note(self):
        return self(**{key: val.to_extension_note(self) for key, val in self.items()}, tags=set(self.tags))

    def to_chord_note(self):
        return self(**{key: val.to_chord_note(self) for key, val in self.items()}, tags=set(self.tags))

    def to_standard_note(self):
        return self(**{key: val.to_standard_note(self) for key, val in self.items()}, tags=set(self.tags))


    def o_melody(self, octave):
        """Change the octave of the melody (not the chord)

        Parameters
        ----------
        octave : int,
                Number of octaves

        Returns
        -------
        chord : Chord

        """
        new_parts = {}
        for part in self.score.keys():
            new_parts[part] = self.score[part].o(octave)

        return self(**new_parts).add_tags(self.tags)

    def patternize(self,
                 nb_excluded_instruments=0,
                 fixed_bass=True,
                 voice_leading=True,
                 melody=False,
                 instruments=None,
                 voicing=None,
                 add_metadata=True,
                   max_duration=64,
                   **kwargs):
        """
        Extract the pattern from the chord

        Parameters
        ----------
        nb_excluded_instruments: int (Default value = 0)
            Number of instruments to exclude from the pattern
        fixed_bass: bool (Default value = True)
            If True, the bass will be fixed in the pattern response
        voice_leading: bool (Default value = True)
            If True, the voice leading will be applied in the pattern response
        melody: bool (Default value = False)
            Do you want to extract a melody or an accompaniment pattern
        instruments: list[str] (Default value = None)
            List of instruments to use for the pattern
        voicing: list[Note] or None (Default value = None)
            Voicing to use for the pattern, otherwise extract the good one
        add_metadata: bool (Default value = True)
            If True, add metadata and features to the pattern
        max_duration: int (Default value = 64)
            Maximum duration of the chord to apply patternize
        Returns
        -------
        score_pattern: Chord
            Patternized chord or score
        pattern: dict
            Pattern data of the score

        """
        from musiclang.analyze.pattern_analyzer import PatternExtractor
        if self.duration > max_duration:
            raise Exception(f'Patternize only works on chords with duration <= {max_duration} increase the max_duration parameter if you want to use it on longer chords')

        dict_pattern = PatternExtractor(
                 nb_excluded_instruments=nb_excluded_instruments,
                 fixed_bass=fixed_bass,
                 voice_leading=voice_leading,
                 melody=melody,
                 instruments=instruments,
                 voicing=voicing
                 ).extract(self)

        pattern = dict_pattern['orchestra']['pattern']
        if add_metadata:
            from musiclang.analyze.pattern_analyzer import PatternFeatureExtractor
            dict_pattern = PatternFeatureExtractor().extract(dict_pattern,
                                                             self,
                                                             melody=melody,
                                                             nb_excluded_instruments=nb_excluded_instruments,
                                                             )
        dict_pattern['multi_chord'] = False
        return dict_pattern, pattern

    def project_pattern(self, score, restart_each_chord=False):
        return self.to_score().project_pattern(score, restart_each_chord=restart_each_chord)

    def remove_drums(self):
        return self(**{ins: val for ins, val in self.score.items() if not ins.startswith('drums')})


    def o(self, octave):
        """Chord up or down the amount of octave in parameter, it will change the chord octave, not the melody

        Parameters
        ----------
        octave :
            return:

        Returns
        -------

        """
        c = self.copy()
        c.octave += octave
        return c


    def delete_instruments(self, instruments):
        return self(**{key: val for key, val in self.score.items() if key not in instruments})

    def delete_instrument_names(self, instrument_names):
        return self(**{key: val for key, val in self.score.items() if key.split('__')[0] not in instrument_names})

    def get_chord_between(self, start, end, complete_if_missing=False):
        """
        Parameters
        ----------
        start :
            
        end :

        complete_if_missing:
            (Default value = False)
        Complete chord with silence if missing time in chord

        Returns
        -------

        """
        from .time_utils import get_chord_between
        return get_chord_between(self, start, end, complete_if_missing=complete_if_missing)

    def copy(self):
        """
        Copy the current chord
        """
        return Chord(element=self.element,
                     extension=self.extension,
                     tonality=self.tonality.copy() if self.tonality is not None else None,
                     score={k: s.copy() for k, s in self.score.items() if s is not None} if self.score is not None else None,
                     octave=self.octave,
                     tags=set(self.tags)
                     )




    def preparse_named_melodies(self, named_melodies):
        """
        Helper function to deal with dictionary of melodies that can contain a list of melodies for some parts.
        It will automatically convert drums melodies to drums track

        Parameters
        ----------
        named_melodies : dict
                         The dictionary of named melodies
            

        Returns
        -------
        res: dict
             The

        Examples
        --------
        >>> from musiclang.library import *
        >>> from musiclang import Chord
        >>> D = {"piano":[s0, s2, s4.e + s2.e]}
        >>> Chord.preparse_named_melodies(D)
        {
            piano__0:s0,
            piano__1:s2,
            piano__2:s4.e + s2.e
            }

        """
        from musiclang import Melody

        named_melodies_result = {}
        for key in named_melodies.keys():
            key_obj = key.split('__')
            if len(key_obj) == 1:
                key_obj = key_obj[0]
                number = 0
            else:
                key_obj, number = key_obj[0], key_obj[1]
                number = int(number)

            if (isinstance(named_melodies[key], list)):
                for idx, melody in enumerate(named_melodies[key]):
                    mel = melody.to_melody()
                    if key_obj.startswith('drums'):
                        new_mel = None
                        for n in mel.notes:
                            new_mel += n.convert_to_drum_note(self)
                        mel = new_mel
                    named_melodies_result[key_obj + '__' + str(number + idx)] = mel
            else:
                mel = named_melodies[key].to_melody()
                if key_obj.startswith('drums'):
                    new_mel = None
                    for n in mel.notes:
                        new_mel += n.convert_to_drum_note(self)
                    mel = new_mel
                named_melodies_result[key_obj + '__' + str(number)] = mel

        return named_melodies_result

    def items(self):
        return self.score.items()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    def augment(self, duration):
        from musiclang import Silence
        if self.empty_score:
            silence = Silence(duration)
            return self(silence)
        else:
            return self(**{part: melody.augment(duration) for part, melody in self.score.items()}, tags=self.tags)

    def set_duration(self, duration):
        from musiclang import Silence
        if self.empty_score:
            if isinstance(duration, float):
                duration = frac(duration).limit_denominator(8)
            silence = Silence(duration)
            return self(silence)
        else:
            return self(**{part: melody.set_duration(duration) for part, melody in self.score.items()}, tags=self.tags)

    def __getattr__(self, item):
        if item in STR_TO_DURATION.keys() and self.empty_score:
            from musiclang import Silence
            duration = STR_TO_DURATION[item]
            silence = Silence(duration)
            return self(silence)

        if len(self.parts) == 0:
            raise AttributeError()
        try:
            res = self(**{part: getattr(self.score[part], item) for part in self.parts}, tags=self.tags)
            return res
        except Exception:
            raise AttributeError(f"Not existing attribute {item}")

    def __call__(self, *melodies, tags=None, **named_melodies):
        """
        Method that allows to assign melodies and instruments to a chord

        Parameters
        ----------
        *melodies: List[Melody]
                   Melodies that will be assigned to default instrument (piano)
        **named_melodies: Dict[str, Melody]
                        Dictionary of named melodies that will be unpacked if necessary

        Returns
        -------

        Examples
        ________

        >>> from musiclang.library import *
        >>> score =  (I % I.M)(s0, s2)
        >>> score
        (I % I.M)(
            piano_0=s0,
            piano_1=s2)

        You can also pass a list of melodies to an instrument :

        >>> score =  (I % I.M)(violin=[s0, s2, s4.e + s2.e])
        >>> score
        (I % I.M)(
            violin__0=s0,
            violin__1=s2,
            violin__2=s4.e + s2.e)

        Finally can also pass a list of voices to an instrument :

        >>> score =  (I % I.M)(piano__0=s0 + s2, piano__2= s2 + s4)
        >>> score
         (I % I.M)(
             piano__0=s0 + s2,
             piano__2= s2 + s4)
        """
        chord = self.copy()
        if tags is not None:
            chord = chord.add_tags(tags)
        named_melodies_preparsed = self.preparse_named_melodies(named_melodies)
        chord.score = {**{f'piano__{i}': melody.to_melody() for i, melody in enumerate(melodies)}, **named_melodies_preparsed}
        return chord

    def melody_to_str(self):
        """
        Convert The chord score to its string representation

        Returns
        -------
        res: str

        """
        return '\n' + ', \n'.join([f"\t{k}=" + str(melody) for k, melody in self.score.items()])

    def __repr__(self):
        return f"{self.to_code()}({self.melody_to_str()})"

    @property
    def full_octave(self):
        return self.octave + self.tonality.octave


    def replace_instruments_names(self, **instruments_dict):
        """
        Replace any instrument with another (use part name (eg: piano__0)

        Parameters
        ----------
        instruments_dict: dict[str, str]
            Dictionary of parts name to replace

        Returns
        -------
        chord: Chord

        """
        new_chord_dict = {}
        idx_instruments = {}
        instruments = sorted(self.score.keys(), key=lambda x: (x.split('__')[0], int(x.split('__')[1])))
        for ins in instruments:
            name, idx = ins.split('__')
            idx = int(idx)
            new_name = instruments_dict.get(name, name)
            if new_name not in idx_instruments.keys():
                idx_instruments[new_name] = -1
            chosen_idx = idx_instruments[new_name] + 1 if idx <= idx_instruments[new_name] else idx
            new_chord_dict[new_name + '__' + str(chosen_idx)] = self.score[ins]
            idx_instruments[new_name] = max(idx, idx_instruments[new_name])

        return self(**new_chord_dict)

    def replace_instruments(self, **instruments_dict):
        """
        Replace any instrument with another (use full part name (eg: piano__0)

        Parameters
        ----------
        instruments_dict: dict[str, str]
            Dictionary of parts name to replace

        Returns
        -------
        chord: Chord

        """
        instruments_dict = {key: instruments_dict[key] if key in instruments_dict.keys() else key for key in self.parts}
        new_chord_dict = {}
        for ins_name, new_ins_name in instruments_dict.items():
            if ins_name in self.score.keys():
                new_chord_dict[new_ins_name] = self.score[ins_name]

        return self(**new_chord_dict)


    def normalize_extension(self):
        extension, replacements, additions, removals = self.get_extension_properties()
        replacements = ''.join([f'({r})' for r in replacements])
        additions = ''.join([f'[{a}]' for a in additions])
        removals = ''.join(['{' + r + '}' for r in removals])
        return extension + replacements + additions + removals

    def get_extension_properties(self):
        extension = self.extension.split('|')[0]
        replacements = sorted(re.findall(r'\((.*?)\)', extension))
        additions = sorted(re.findall(r'\[(.*?)\]', extension))
        removals = sorted(re.findall(r'\{(.*?)\}', extension))
        ext = extension
        for r in replacements + additions + removals:
            ext = ext.replace(r, '')
        ext = ext.replace('()', '').replace('[]', '').replace('{}', '')
        extension = ext
        return extension, replacements, additions, removals


    def _chord_notes_calc(self, extension, replacements, additions, removals):
        from .library import BASE_EXTENSION_DICT, \
            DICT_REPLACEMENT, DICT_ADDITION, DICT_REMOVAL
        notes = BASE_EXTENSION_DICT[extension][:]
        notes_without_octave = [n.o(-n.octave) for n in notes]

        dict_replaced = {}
        for replacement in replacements:
            note_replaced, new_note = DICT_REPLACEMENT[replacement]
            if note_replaced not in notes_without_octave:
                additions.append(replacement)
            else:
                idx = notes_without_octave.index(note_replaced)
                notes[idx] = new_note.o(notes[idx].octave)
                notes_without_octave[idx] = new_note.o(-new_note.octave)
                dict_replaced[note_replaced] = notes_without_octave[idx]

        for addition in additions:
            note_after, new_note = DICT_ADDITION[addition]
            if note_after in dict_replaced.keys():
                query_note = dict_replaced[note_after]
            else:
                query_note = note_after
            idx = notes_without_octave.index(query_note) + 1
            new_note = new_note.o(note_after.octave)
            notes.insert(idx, new_note)
            notes_without_octave.insert(idx, new_note.o(-new_note.octave))

        for removal in removals:
            note_removed = DICT_REMOVAL[removal]
            idx = len(notes) - notes_without_octave[::-1].index(note_removed) - 1
            notes.pop(idx)
            notes_without_octave.pop(idx)
        # Sort per pitch
        notes = list(sorted(notes, key=lambda x: self.to_pitch(x)))
        return notes

    @property
    def chord_notes(self):
        extension, replacements, additions, removals = self.get_extension_properties()
        new_extension = ''
        if extension in ['2', '65', '43', '7']:
            new_extension = '7'
        elif extension == '9':
            new_extension = '9'
        elif extension == '11':
            new_extension = '11'
        elif extension == '13':
            new_extension = '13'

        return self._chord_notes_calc(new_extension, replacements, additions, removals)

    @cached_property
    def extension_notes(self):
        # Try a composite with a match
        extension, replacements, additions, removals = self.get_extension_properties()
        return self._chord_notes_calc(extension, replacements, additions, removals)


    def to_voicing(self, nb_voices=4, instruments=None):
        """Convert score to a four voice voicing using the extensions provided in the chord.

        It will remove the existing scores of each chord and create the associated voicings

        Parameters
        ----------
        instruments : None or List[str] (Default value = ['piano__0', 'piano__1', 'piano__2', 'piano__3'])
                      The list of instruments used to create the voicings.

        Returns
        -------

        score: Score
               The score with voicings corresponding to chords

        """

        if instruments is None:
            instruments = [f'piano__{i}' for i in range(nb_voices)]

        notes = self.extension_notes[:nb_voices]
        if not self.empty_score:
            notes = [n.set_duration(self.duration) for n in notes]
        if len(notes) < nb_voices:
            # Add to match octaves
            for i in range(len(notes), nb_voices):
                n = notes[i % len(notes)].copy()
                oct = (i // len(notes))
                n.octave += oct
                notes.append(n)
        chord = self(**{ins: notes[i].copy() for i, ins in enumerate(instruments)})

        return chord


    @property
    def empty_score(self):
        return len(self.score.keys()) == 0

    def to_code(self):
        """Export the code as python interpretable str.
        It ignores the score
        :return:

        Parameters
        ----------

        Returns
        -------

        """
        chord_str = f"({self.element_to_str()}{self.extension_to_str()}{self.tonality_to_str()})"
        if self.octave != 0:
            chord_str = f"{chord_str}.o({self.octave})"

        return chord_str

    def to_midi(self, filepath, **kwargs):
        """
        Save the chord to midi

        See Also
        --------

        :func:`~Score.to_midi()`

        Parameters
        ----------
        filepath :
            
        **kwargs :
            

        Returns
        -------

        """
        from .score import Score
        return Score([self]).to_midi(filepath, **kwargs)


    def to_scale_notes(self):
        """
        Convert the notes of a chord in scale notes

        Parameters
        ----------

        Returns
        -------

        """
        return self(**{ins: n.to_scale_notes(self) for ins, n in self.score.items()})

    def project_on_rhythm(self, rhythm, **kwargs):
        return self(**{ins: n.project_on_rhythm(rhythm, chord=self, **kwargs) for ins, n in self.score.items()})

    def set_degree(self, degree):
        res = self.copy()
        res.element = degree
        return res

    def set_octave(self, octave):
        res = self.copy()
        res.octave = octave
        return res
    def set_element(self, element):
        return self.set_degree(element)

    def to_custom_chords(self, nb_voices=4):
        return self.to_score().to_custom_chords(nb_voices=nb_voices)


    def to_orchestra(self, drop_drums=True, nb=4, pattern=False):
        from musiclang.transform.composing.chord_to_orchestra import chord_to_orchestra
        return chord_to_orchestra(self, drop_drums=drop_drums, nb=nb, pattern=pattern)

    def repeated_notes_to_legato(self):
        return self(**{ins: n.repeated_notes_to_legato() for ins, n in self.score.items()})

    def get_orchestration(self):
        from .melody import Melody
        def voicing_to_quality(voicing, chord):
            notes = chord.extension_notes
            length = len(notes)
            length = min(4, length)
            qualities = []

            data_dict = {
                (0, 1): 'lowest_note',
                (0, 2): 'lowest_note',
                (0, 3): 'lowest_note',
                (0, 4): 'lowest_note',

                (1, 1): 'lowest_note',
                (1, 2): 'top_note',
                (1, 3): 'middle_note1',
                (1, 4): 'middle_note1',

                (2, 1): 'lowest_note',
                (2, 2): 'lowest_note',
                (2, 3): 'top_note',
                (2, 4): 'middle_note2',

                (3, 1): 'lowest_note',
                (3, 2): 'top_note',
                (3, 3): 'lowest_note',
                (3, 4): 'top_note',

            }
            for note in voicing:
                octave = note.octave
                value = note.val
                quality = data_dict[(value, length)]
                qualities.append((quality, octave))

            return qualities

        orchestration = []
        data, pattern = self.patternize()
        instruments = data['instruments']
        voicing = data['voicing']
        qualities = voicing_to_quality(voicing, data['orchestra']['pattern'])

        instruments_idx = {}
        for idx, (ins, quality) in enumerate(zip(instruments, qualities)):
            note, octave = quality
            local_pattern = pattern.score[f'v__{idx}']
            local_pattern = Melody([n if n.type != 'x' else n.set_val(0) for n in local_pattern.notes])
            rhythm = local_pattern.quantize_melody()
            instruments_idx[ins] = instruments_idx.get(ins, -1) + 1
            orchestration.append({'instrument': ins, 'octave': octave,
                                  'note': note, 'pattern': local_pattern,
                                  'rhythm': rhythm
                                  })

        return orchestration

    @classmethod
    def from_romantext(self, text, time_signature, tonality):
        from musiclang import ScoreFormatter
        full_text = f'Time Signature: {"/".join([str(t) for t in time_signature])}'
        full_text += f'\nTonality: {tonality}'
        full_text += f'\nm0 {text}'
        return ScoreFormatter(full_text).parse()[0]

    def to_romantext(self, tonality):

        def to_degree_type(degree, mode):
            L = {
                'm': ['i', 'ii%', 'III+', 'iv', 'V', 'VI', 'viio'],
                'M': ['I', 'ii', 'iii', 'IV', 'V', 'vi', 'vii%']
            }
            return L[mode][degree]

        tone = tonality.degree
        mode = tonality.mode

        extension_tonality = ''
        if mode != self.tonality.mode or tone != self.tonality.degree:
            extension_tonality = f"/{self.tonality.to_romantext(tonality)}"

        # Now find the degree
        degree = to_degree_type(self.degree, self.tonality.mode)

        # Find proper extension
        extension = self.extension

        # Regex to list everything between "{}" and everything between "[]"
        import re
        reg_omit = r"\{(.*?)\}"
        reg_add = r"\[(.*?)\]"

        matches_omit = re.findall(reg_omit, extension)
        matches_add = re.findall(reg_add, extension)

        to_remove = ["{" + f"{match}" + "}" for match in matches_omit]
        to_remove += ["[" + f"{match}" + "]" for match in matches_add]

        formatted_extension = extension
        for r in to_remove:
            formatted_extension = formatted_extension.replace(r, '')

        # Write a script that matches the text which is not between {} or []
        # For example "64{-5}[add6][add2]" should return ["64"]

        for match in matches_omit:
            formatted_extension += f"[{match.replace('-', 'no')}]"

        for match in matches_add:
            formatted_extension += f"[{match}]"

        formatted_extension = formatted_extension.replace('(+)', '+')

        return f"{degree}{formatted_extension}{extension_tonality}"

